#!/usr/bin/env python3.5

'''

purl - python curl module

'''


# import  threading
import  argparse
import  time
import  sys
import  json
import  pprint
import  socket
import  pycurl
import  io
import  os
import  urllib
import  datetime
import  codecs

import  pudb

# pman local dependencies
from    ._colors         import  Colors
from    .pfioh           import  *
from    .message         import  Message

# A global variable that tracks if script was started from CLI or programmatically
b_startFromCLI  = False

class Purl():

    ''' Represents an example client. '''

    def qprint(self, msg, **kwargs):

        str_comms  = "status"
        for k,v in kwargs.items():
            if k == 'comms':    str_comms  = v

        if self.b_useDebug:
            write   = self.debug
        else:
            write   = print

        # pudb.set_trace()

        if not self.b_quiet:
            if not self.b_useDebug:
                if str_comms == 'status':   write(Colors.PURPLE,    end="")
                if str_comms == 'error':    write(Colors.RED,       end="")
                if str_comms == "tx":       write(Colors.YELLOW + "---->")
                if str_comms == "rx":       write(Colors.GREEN  + "<----")
                write('%s' % datetime.datetime.now() + " ",       end="")
            write(' | ' + msg)
            if not self.b_useDebug:
                if str_comms == "tx":       write(Colors.YELLOW + "---->")
                if str_comms == "rx":       write(Colors.GREEN  + "<----")
                write(Colors.NO_COLOUR, end="")

    def col2_print(self, str_left, str_right):
        print(Colors.WHITE +
              ('%*s' % (self.LC, str_left)), end='')
        print(Colors.LIGHT_BLUE +
              ('%*s' % (self.RC, str_right)) + Colors.NO_COLOUR)

    def __init__(self, **kwargs):
        # threading.Thread.__init__(self)

        self._log                   = Message()
        self._log._b_syslog         = True
        self.__name                 = "Purl"
        self.b_useDebug             = False
        self._startFromCLI          = False

        str_debugDir                = '%s/tmp' % os.environ['HOME']
        if not os.path.exists(str_debugDir):
            os.makedirs(str_debugDir)
        self.str_debugFile          = '%s/debug-charm.log' % str_debugDir

        self.str_http           = ""
        self.str_ip             = ""
        self.str_port           = ""
        self.str_URL            = ""
        self.str_verb           = ""
        self.str_msg            = ""
        self.str_auth           = ""
        self.d_msg              = {}
        self.str_protocol       = "http"
        self.pp                 = pprint.PrettyPrinter(indent=4)
        self.b_man              = False
        self.str_man            = ''
        self.b_quiet            = False
        self.b_raw              = False
        self.b_oneShot          = False
        self.auth               = ''
        self.str_jsonwrapper    = ''
        self.str_contentType    = ''
        self.b_useDebug         = False
        self.str_debugFile      = ''

        self.LC                 = 40
        self.RC                 = 40

        for key,val in kwargs.items():
            if key == 'msg':
                self.str_msg                = val
                try:
                    self.d_msg              = json.loads(self.str_msg)
                except:
                    pass
            if key == 'http':           self.httpStr_parse( http    = val)
            if key == 'auth':           self.str_auth               = val
            if key == 'verb':           self.str_verb               = val
            if key == 'contentType':    self.str_contentType        = val
            if key == 'ip':             self.str_ip                 = val
            if key == 'port':           self.str_port               = val
            if key == 'b_quiet':        self.b_quiet                = val
            if key == 'b_raw':          self.b_raw                  = val
            if key == 'b_oneShot':      self.b_oneShot              = val
            if key == 'man':            self.str_man                = val
            if key == 'jsonwrapper':    self.str_jsonwrapper        = val
            if key == 'useDebug':       self.b_useDebug             = val
            if key == 'debugFile':      self.str_debugFile          = val
            if key == 'startFromCLI':   self._startFromCLI          = val

        if self.b_useDebug:
            self.debug                  = Message(logTo = self.str_debugFile)
            self.debug._b_syslog        = True
            self.debug._b_flushNewLine  = True

        if len(self.str_man):
            print(self.man(on = self.str_man))
            sys.exit(0)

        if not self.b_quiet:

            print(Colors.LIGHT_GREEN)
            print("""
            \t\t\t\t+--------------------+
            \t\t\t\t|  Welcome to purl!  |
            \t\t\t\t+--------------------+
            """)
            print(Colors.CYAN + """
            'purl' is a wrapper about pycurl, originally designed for the ChRIS/pman interface.
            It also functions as a reasonable CLI curl client, too!

            Type 'purl --man commands' for more help.""")
            if self.b_useDebug:
                print("""
            Debugging output is directed to the file '%s'.
                """ % (self.str_debugFile))
            else:
                print("""
            Debugging output will appear in *this* console.
                """)

            self.qprint('purl: Start from CLI = %d' % self._startFromCLI)
            self.qprint('purl: Command line args = %s' % sys.argv)
            if self._startFromCLI and (sys.argv) == 1: sys.exit(1)

            self.col2_print("Will transmit to",     '%s://%s:%s' % (self.str_protocol, self.str_ip, self.str_port))


    def man(self, **kwargs):
        """
        Print some man for each understood command
        """

        str_man     = 'commands'
        str_amount  = 'full'

        for k, v in kwargs.items():
            if k == 'on':       str_man     = v
            if k == 'amount':   str_amount  = v

        if str_man == 'commands':
            str_commands = """
            This script/module provides CURL-based GET/PUT/POST communication over http
            to a remote REST-like service: """ + Colors.GREEN + """

                 ./purl.py [--auth <username:passwd>] [--verb <GET/POST>]   \\
                            --http <IP>[:<port>]</some/path/>

            """ + Colors.WHITE + """
            Where --auth is an optional authorization to pass to the REST API,
            --verb denotes the REST verb to use and --http specifies the REST URL.

            Additionally, a 'message' described in JSON syntax can be pushed to the
            remote service, in the following syntax: """ + Colors.GREEN + """

                 ./purl.py [--auth <username:passwd>] [--verb <GET/POST>]   \\
                            --http <IP>[:<port>]</some/path/>               \\
                           [--msg <JSON-formatted-string>]

            """ + Colors.WHITE + """
            In the case of the 'pman' system this --msg flag has very specific
            contextual syntax, for example:
            """ + Colors.GREEN + """
                 ./purl.py --verb POST --http %s:%s/api/v1/cmd/ --msg \\
                                '{  "action": "run",
                                    "meta": {
                                        "cmd":      "cal 7 1970",
                                        "auid":     "rudolphpienaar",
                                        "jid":      "<jid>-1",
                                        "threaded": true
                                    }
                                }'


            """ % (self.str_ip, self.str_port) + Colors.CYAN + """

            The following specific action directives are directly handled by script:
            """ + "\n" + \
            self.man_pushPath(          description =   "short")      + "\n" + \
            self.man_pullPath(          description =   "short")      + "\n" + \
            Colors.YELLOW + \
            """
            To get detailed help on any of the above commands, type
            """ + Colors.LIGHT_CYAN + \
            """
                ./purl.py --man <command>
            """

            return str_commands

        if str_man  == 'pushPath':  return self.man_pushPath(       description  =   str_amount)
        if str_man  == 'pullPath':  return self.man_pullPath(       description  =   str_amount)

    def man_pushPath(self, **kwargs):
        """
        """

        b_fullDescription   = False
        str_description     = "full"

        for k,v in kwargs.items():
            if k == 'description':  str_description = v
        if str_description == "full":   b_fullDescription   = True

        str_manTxt =   Colors.LIGHT_CYAN        + \
                       "\t\t%-20s" % "pushPath" + \
                       Colors.LIGHT_PURPLE      + \
                       "%-60s" % "push a filesystem path over HTTP." + \
                       Colors.NO_COLOUR

        if b_fullDescription:
            str_manTxt += """

                This pushes a file over HTTP. The 'meta' dictionary
                can be used to specify content specific information
                and other information.

                Note that the "file" server is typically *not* on the
                same port as the `pman` process. Usually a prior call
                must be made to `pman` to start a one-shot listener
                on a given port. This port then accepts the file transfer
                from the 'pushPath' method.
                
                The "meta" dictionary consists of several nested 
                dictionaries. In particular, the "remote/path"
                field can be used to suggest a location on the remote
                filesystem to save the transmitted data. Successful
                saving to this path depends on whether or not the
                remote server process actually has permission to
                write in that location.

                """ + Colors.YELLOW + """EXAMPLE:
                """ + Colors.LIGHT_GREEN + """
                purl --verb POST --http %s:%s/api/v1/cmd/ --msg \\
                    '{  "action": "pushPath",
                        "meta":
                            {
                                "local":
                                    {
                                        "path":         "/path/on/client"
                                    },
                                "remote":
                                    {
                                        "path":         "/path/on/server"
                                    },
                                "transport":
                                    {
                                        "mechanism":    "compress",
                                        "compress": {
                                            "encoding": "base64",
                                            "archive":  "zip",
                                            "unpack":   true,
                                            "cleanup":  true
                                        }
                                    }
                            }
                    }'
                """ % (self.str_ip, self.str_port) + Colors.NO_COLOUR  + """
                """ + Colors.YELLOW + """ALTERNATE -- using copy/symlink:
                """ + Colors.LIGHT_GREEN + """
                purl --verb POST --http %s:%s/api/v1/cmd/ --msg \\
                    '{  "action": "pushPath",
                        "meta":
                            {
                                "local":
                                    {
                                        "path":         "/path/on/client"
                                    },
                                "remote":
                                    {
                                        "path":         "/path/on/server"
                                    },
                                "transport":
                                    {
                                        "mechanism":    "copy",
                                        "copy": {
                                            "symlink": true
                                        }
                                    }
                            }
                    }'
                """ % (self.str_ip, self.str_port) + Colors.NO_COLOUR

        return str_manTxt

    def man_pullPath(self, **kwargs):
        """
        """

        b_fullDescription   = False
        str_description     = "full"

        for k,v in kwargs.items():
            if k == 'description':  str_description = v
        if str_description == "full":   b_fullDescription   = True

        str_manTxt =   Colors.LIGHT_CYAN        + \
                       "\t\t%-20s" % "pullPath" + \
                       Colors.LIGHT_PURPLE      + \
                       "%-60s" % "pull a filesystem path over HTTP." + \
                       Colors.NO_COLOUR

        if b_fullDescription:
            str_manTxt += """

                This pulls data over HTTP from a remote server.
                The 'meta' dictionary can be used to specify content
                specific information and other detail.

                Note that the "file" server is typically *not* on the
                same port as a `pman` process. Usually a prior call
                must be made to `pman` to start a one-shot listener
                on a given port. This port then accepts the file transfer
                from the 'pullPath' method.

                The "meta" dictionary consists of several nested
                dictionaries. In particular, the "remote/path"
                field can be used to specify a location on the remote
                filesystem to pull. Successful retrieve from this path
                depends on whether or not the remote server process actually
                has permission to read in that location.

                """ + Colors.YELLOW + """EXAMPLE -- using zip:
                """ + Colors.LIGHT_GREEN + """
                purl --verb POST --http %s:%s/api/v1/cmd/ --msg \\
                    '{  "action": "pullPath",
                        "meta":
                            {
                                "local":
                                    {
                                        "path":         "/path/on/client"
                                    },
                                "remote":
                                    {
                                        "path":         "/path/on/server"
                                    },
                                "transport":
                                    {
                                        "mechanism":    "compress",
                                        "compress": {
                                            "encoding": "base64",
                                            "archive":  "zip",
                                            "unpack":   true,
                                            "cleanup":  true
                                        }
                                    }
                            }
                    }'
                """ % (self.str_ip, self.str_port) + Colors.NO_COLOUR + """
                """ + Colors.YELLOW + """ALTERNATE -- using copy/symlink:
                """ + Colors.LIGHT_GREEN + """
                purl --verb POST --http %s:%s/api/v1/cmd/ --msg \\
                    '{  "action": "pullPath",
                        "meta":
                            {
                                "local":
                                    {
                                        "path":         "/path/on/client"
                                    },
                                "remote":
                                    {
                                        "path":         "/path/on/server"
                                    },
                                "transport":
                                    {
                                        "mechanism":    "copy",
                                        "copy": {
                                            "symlink": true
                                        }
                                    }
                            }
                    }'
                """ % (self.str_ip, self.str_port) + Colors.NO_COLOUR

        return str_manTxt

    def pull_core(self, **kwargs):
        """
        Just the core of the pycurl logic.
        """

        str_ip              = self.str_ip
        str_port            = self.str_port
        verbose             = 0
        d_msg               = {}

        for k,v in kwargs.items():
            if k == 'ip':       str_ip      = v
            if k == 'port':     str_port    = v
            if k == 'msg':      d_msg       = v
            if k == 'verbose':  verbose     = v

        response            = io.BytesIO()

        str_query   = ''
        if len(d_msg):
            d_meta              = d_msg['meta']
            str_query           = '?%s' % urllib.parse.urlencode(d_msg)

        str_URL = "http://%s:%s%s%s" % (str_ip, str_port, self.str_URL, str_query)

        self.qprint(str_URL,
                    comms  = 'tx')

        c                   = pycurl.Curl()
        c.setopt(c.URL, str_URL)
        if verbose: c.setopt(c.VERBOSE, 1)
        c.setopt(c.FOLLOWLOCATION,  1)
        c.setopt(c.WRITEFUNCTION,   response.write)
        if len(self.str_auth):
            c.setopt(c.USERPWD, self.str_auth)
        self.qprint("Waiting for PULL response...", comms = 'status')
        c.perform()
        c.close()
        try:
            str_response        = response.getvalue().decode()
        except:
            str_response        = response.getvalue()

        self.qprint('Incoming transmission received, length = %s' % "{:,}".format(len(str_response)),
                    comms = 'rx')
        return str_response

    def pullPath_core(self, d_msg, **kwargs):
        """
        Just the core of the pycurl logic.
        """

        str_response = self.pull_core(msg = self.d_msg)

        if len(str_response) < 800:
            # It's possible an error occurred for the response to be so short.
            # Try and json load, and examine for 'status' field.
            b_response      = False
            b_status        = False
            try:
                d_response  = json.loads(str_response)
                b_response  = True
                b_status    = d_response['status']
                str_error   = d_response
            except:
                str_error   = str_response
            if not b_status or 'Network Error' in str_response:
                self.qprint('Some error occurred at remote location:',
                            comms = 'error')
                return {'status':       False,
                        'msg':          'PULL unsuccessful',
                        'response':     str_error,
                        'timestamp':    '%s' % datetime.datetime.now(),
                        'size':         "{:,}".format(len(str_response))}
            else:
                return {'status':       d_response['status'],
                        'msg':          'PULL successful',
                        'response':     d_response,
                        'timestamp':    '%s' % datetime.datetime.now(),
                        'size':         "{:,}".format(len(str_response))}

        self.qprint("Received " + Colors.YELLOW + "{:,}".format(len(str_response)) +
                    Colors.PURPLE + " bytes..." ,
                    comms = 'status')

        return {'status':       True,
                'msg':          'PULL successful',
                'response':     str_response,
                'timestamp':    '%s' % datetime.datetime.now(),
                'size':         "{:,}".format(len(str_response))}

    def pullPath_compress(self, d_msg, **kwargs):
        """

        This pulls a compressed path from a remote host/location.

        """

        # Parse "header" information
        d_meta                  = d_msg['meta']
        d_local                 = d_meta['local']
        str_localPath           = d_local['path']
        d_remote                = d_meta['remote']
        d_transport             = d_meta['transport']
        d_compress              = d_transport['compress']
        d_ret                   = {}
        d_ret['remoteServer']   = {}
        d_ret['localOp']        = {}

        if 'cleanup' in d_compress:
            b_cleanZip      = d_compress['cleanup']

        # Pull the actual data into a dictionary holder
        d_pull                  = self.pullPath_core(d_msg)
        d_ret['remoteServer']   = d_pull

        if not d_pull['status']:
            return {'stdout': json.dumps(d_pull['stdout'])}

        str_localStem       = os.path.split(d_remote['path'])[-1]
        str_fileSuffix      = ""
        if d_compress['archive']     == "zip":       str_fileSuffix   = ".zip"

        str_localFile       = "%s/%s%s" % (d_meta['local']['path'], str_localStem, str_fileSuffix)
        str_response        = d_pull['response']
        d_pull['response']  = '<truncated>'

        if d_compress['encoding'] == 'base64':
            self.qprint("Decoding base64 encoded text stream to %s..." % \
                        str_localFile, comms = 'status')
            d_fio = base64_process(
                action          = 'decode',
                payloadBytes    = str_response,
                saveToFile      = str_localFile
            )
            d_ret['localOp']['decode']   = d_fio
        else:
            self.qprint("Writing byte stream to %s..." % str_localFile,
                        comms = 'status')
            with open(str_localFile, 'wb') as fh:
                fh.write(str_response)
                fh.close()
            d_ret['localOp']['stream']                  = {}
            d_ret['localOp']['stream']['status']        = True
            d_ret['localOp']['stream']['fileWritten']   = str_localFile
            d_ret['localOp']['stream']['timestamp']     = '%s' % datetime.datetime.now()
            d_ret['localOp']['stream']['filesize']      = "{:,}".format(len(str_response))

        if d_compress['archive'] == 'zip':
            self.qprint("Unzipping %s to %s"  % (str_localFile, str_localPath),
                        comms = 'status')
            d_fio = zip_process(
                action          = "unzip",
                payloadFile     = str_localFile,
                path            = str_localPath
            )
            d_ret['localOp']['unzip']       = d_fio
            d_ret['localOp']['unzip']['timestamp']  = '%s' % datetime.datetime.now()
            d_ret['localOp']['unzip']['filesize']   = '%s' % "{:,}".format(os.stat(d_fio['fileProcessed']).st_size)
            d_ret['status']                 = d_fio['status']
            d_ret['msg']                    = d_fio['msg']

        if b_cleanZip and d_ret['status']:
            self.qprint("Removing zip file %s..." % str_localFile,
                        comms = 'status')
            os.remove(str_localFile)

        return d_ret

    def pullPath_copy(self, d_msg, **kwargs):
        """
        Handle the "copy" pull operation
        """

        # Parse "header" information
        d_meta              = d_msg['meta']
        d_local             = d_meta['local']
        str_localPath       = d_local['path']
        d_remote            = d_meta['remote']
        d_transport         = d_meta['transport']
        d_copy              = d_transport['copy']

        # Pull the actual data into a dictionary holder
        d_curl                      = {}
        d_curl['remoteServer']      = self.pullPath_core(d_msg)
        d_curl['copy']              = {}
        d_curl['copy']['status']    = d_curl['remoteServer']['status']
        if not d_curl['copy']['status']:
            d_curl['copy']['msg']   = "Copy on remote server failed!"
        else:
            d_curl['copy']['msg']   = "Copy on remote server success!"

        return d_curl

    def path_remoteLocationCheck(self, d_msg, **kwargs):
        """
        This method checks if the "remote" path is valid.
        """

        # Pull the actual data into a dictionary holder
        d_pull = self.pullPath_core(d_msg)
        return d_pull

    def path_localLocationCheck(self, d_msg, **kwargs):
        """
        Check if a path exists on the local filesystem

        :param self:
        :param kwargs:
        :return:
        """
        d_meta              = d_msg['meta']
        d_local             = d_meta['local']

        str_localPath       = d_local['path']

        b_isFile            = os.path.isfile(str_localPath)
        b_isDir             = os.path.isdir(str_localPath)
        b_exists            = os.path.exists(str_localPath)

        d_ret               = {
            'status':  b_exists,
            'isfile':  b_isFile,
            'isdir':   b_isDir
        }

        return {'check':        d_ret,
                'status':       d_ret['status'],
                'timestamp':    '%s' % datetime.datetime.now()}

    def push_core(self, d_msg, **kwargs):
        """

        """

        str_fileToProcess   = ""
        str_encoding        = "none"
        d_ret               = {}
        str_ip              = self.str_ip
        str_port            = self.str_port
        verbose             = 0

        for k,v in kwargs.items():
            if k == 'fileToPush':   str_fileToProcess   = v
            if k == 'encoding':     str_encoding        = v
            if k == 'd_ret':        d_ret               = v
            if k == 'ip':           str_ip              = v
            if k == 'port':         str_port            = v
            if k == 'verbose':      verbose     = v

        if len(self.str_jsonwrapper):
            str_msg         = json.dumps({self.str_jsonwrapper: d_msg})
        else:
            str_msg         = json.dumps(d_msg)
        response            = io.BytesIO()

        self.qprint("http://%s:%s%s" % (str_ip, str_port, self.str_URL) + '\n '+ str(d_msg),
                    comms  = 'tx')

        c = pycurl.Curl()
        c.setopt(c.POST, 1)
        # c.setopt(c.URL, "http://%s:%s/api/v1/cmd/" % (str_ip, str_port))
        c.setopt(c.URL, "http://%s:%s%s" % (str_ip, str_port, self.str_URL))
        if str_fileToProcess:
            fread               = open(str_fileToProcess, "rb")
            filesize            = os.path.getsize(str_fileToProcess)
            c.setopt(c.HTTPPOST, [  ("local",       (c.FORM_FILE, str_fileToProcess)),
                                    ("encoding",    str_encoding),
                                    ("d_msg",       str_msg),
                                    ("filename",    str_fileToProcess)]
                     )
            c.setopt(c.READFUNCTION,    fread.read)
            c.setopt(c.POSTFIELDSIZE,   filesize)
        else:
            # c.setopt(c.HTTPPOST, [
            #                         ("d_msg",    str_msg),
            #                      ]
            #          )
            c.setopt(c.POSTFIELDS, str_msg)
        if verbose:                     c.setopt(c.VERBOSE, 1)
        # print(self.str_contentType)
        if len(self.str_contentType):   c.setopt(c.HTTPHEADER, ['Content-type: %s' % self.str_contentType])
        c.setopt(c.WRITEFUNCTION,   response.write)
        if len(self.str_auth):
            c.setopt(c.USERPWD, self.str_auth)
        if str_fileToProcess:
            self.qprint("Transmitting " + Colors.YELLOW + "{:,}".format(os.stat(str_fileToProcess).st_size) + \
                        Colors.PURPLE + " bytes...",
                        comms = 'status')
        else:
            self.qprint("Sending data...",
                        comms = 'status')
        try:
            c.perform()
            str_response        = response.getvalue().decode()
        except Exception as e:
            str_exception   = str(e)
            self.qprint('Execption trapped: %s' % str_exception)
            str_response    = str_exception
        c.close()

        self.qprint('response from call = %s' % str_response, comms = 'status')
        if self.b_raw:
            try:
                d_ret           = json.loads(str_response)
            except:
                d_ret           = str_response
        else:
            try:
                d_ret['stdout']     = json.loads(str_response)
            except:
                d_ret['stdout']     = str_response
            if 'status' in d_ret['stdout']:
                d_ret['status']     = d_ret['stdout']['status']
            d_ret['msg']        = 'push OK.'

        if isinstance(d_ret, object):
            self.qprint(json.dumps(d_ret), comms = 'rx')
        if isinstance(d_ret, str):
            self.qprint(d_ret, comms = 'rx')

        return d_ret

    def pushPath_core(self, d_msg, **kwargs):
        """

        """

        str_fileToProcess   = ""
        str_encoding        = "none"
        d_ret               = {}
        for k,v in kwargs.items():
            if k == 'fileToPush':   str_fileToProcess   = v
            if k == 'encoding':     str_encoding        = v
            if k == 'd_ret':        d_ret               = v

        d_meta              = d_msg['meta']
        str_ip              = self.str_ip
        str_port            = self.str_port
        if 'remote' in d_meta:
            d_remote            = d_meta['remote']
            if 'ip' in d_remote:    str_ip      = d_remote['ip']
            if 'port' in d_remote:  str_port    = d_remote['port']

        d_ret               = self.push_core(
                                                fileToPush  = str_fileToProcess,
                                                encoding    = str_encoding,
                                                ip          = str_ip,
                                                port        = str_port
                                            )
        return d_ret

    def pushPath_compress(self, d_msg, **kwargs):
        """
        """

        d_meta              = d_msg['meta']
        str_meta            = json.dumps(d_meta)
        d_local             = d_meta['local']
        str_localPath       = d_local['path']

        d_remote            = d_meta['remote']
        str_ip              = self.str_ip
        str_port            = self.str_port
        if 'ip' in d_remote:
            str_ip          = d_remote['ip']
        if 'port' in d_remote:
            str_port        = d_remote['port']

        str_mechanism       = ""
        str_encoding        = ""
        str_archive         = ""
        d_transport         = d_meta['transport']
        if 'compress' in d_transport:
            d_compress      = d_transport['compress']
            str_archive     = d_compress['archive']
            str_encoding    = d_compress['encoding']

        str_remotePath      = d_remote['path']

        if 'cleanup' in d_compress:
            b_cleanZip      = d_compress['cleanup']


        str_fileToProcess   = str_localPath
        str_zipFile         = ""
        str_base64File      = ""

        b_zip               = True

        if str_archive      == 'zip':   b_zip   = True
        else:                           b_zip   = False

        if os.path.isdir(str_localPath):
            b_zip           = True
            str_archive     = 'zip'

        d_ret               = {}
        d_ret['local']      = {}
        # If specified (or if the target is a directory), create zip archive
        # of the local path
        if b_zip:
            self.qprint("Zipping target...", comms = 'status')
            d_fio   = zip_process(
                action  = 'zip',
                path    = str_localPath,
                arcroot = str_localPath
            )
            if not d_fio['status']: return {'stdout': json.dumps(d_fio)}
            str_fileToProcess   = d_fio['fileProcessed']
            str_zipFile         = str_fileToProcess
            self.qprint("Zipped to %s..." % str_fileToProcess, comms = 'status')
            d_ret['local']['zip']               = d_fio

        # Encode possible binary filedata in base64 suitable for text-only
        # transmission.
        if str_encoding     == 'base64':
            self.qprint("base64 encoding target...", comms = 'status')
            d_fio   = base64_process(
                action      = 'encode',
                payloadFile = str_fileToProcess,
                saveToFile  = str_fileToProcess + ".b64"
            )
            str_fileToProcess       = d_fio['fileProcessed']
            self.qprint("base64 encoded to %s..." % str_fileToProcess, comms = 'status')
            str_base64File          = str_fileToProcess
            d_ret['local']['encoding']                   = d_fio

        # Push the actual file -- note the d_ret!
        d_ret['remoteServer']  = self.push_core(    d_msg,
                                                    fileToPush  = str_fileToProcess,
                                                    encoding    = str_encoding)
                                                    # d_ret       = d_ret)
        d_ret['status'] = d_ret['remoteServer']['status']
        d_ret['msg']    = d_ret['remoteServer']['msg']

        if b_cleanZip:
            self.qprint("Removing temp files...", comms = 'status')
            if os.path.isfile(str_zipFile):     os.remove(str_zipFile)
            if os.path.isfile(str_base64File):  os.remove(str_base64File)

        return d_ret

        # return {'stdout': {'return' : d_ret},
        #         'status': d_ret['fromServer']['status']}

    def pushPath_copy(self, d_msg, **kwargs):
        """
        Handle the "copy" pull operation
        """

        # Parse "header" information
        d_meta              = d_msg['meta']
        d_local             = d_meta['local']
        str_localPath       = d_local['path']
        d_remote            = d_meta['remote']
        d_transport         = d_meta['transport']
        d_copy              = d_transport['copy']

        # Pull the actual data into a dictionary holder
        d_curl                      = {}
        d_curl['remoteServer']      = self.push_core(d_msg)
        d_curl['copy']              = {}
        d_curl['copy']['status']    = d_curl['remoteServer']['status']
        if not d_curl['copy']['status']:
            d_curl['copy']['msg']   = "Copy on remote server failed!"
        else:
            d_curl['copy']['msg']   = "Copy on remote server success!"

        return d_curl

    def pathOp_do(self, d_msg, **kwargs):
        """
        Entry point for path-based push/pull calls.

        Essentially, this method is the central dispatching nexus to various
        specialized push operations.

        """

        d_meta              = d_msg['meta']
        d_transport         = d_meta['transport']
        b_OK                = True
        d_ret               = {}

        str_action          = "pull"
        for k,v, in kwargs.items():
            if k == 'action':   str_action  = v

        # First check on the paths, both local and remote
        self.qprint('Checking local path status...', comms = 'status')
        d_ret['localCheck'] = self.path_localLocationCheck(d_msg)
        if not d_ret['localCheck']['status']:
            self.qprint('An error occurred while checking on the local path.',
                        comms = 'error')
            d_ret['localCheck']['msg']          = 'The local path spec is invalid!'
            d_ret['localCheck']['status'] = False
            b_OK            = False
        else:
            d_ret['localCheck']['msg']          = "Check on local path successful."
        d_ret['status']     = d_ret['localCheck']['status']
        d_ret['msg']        = d_ret['localCheck']['msg']

        if b_OK:
            d_transport['checkRemote']  = True
            self.qprint('Checking remote path status...', comms = 'status')
            d_ret['remoteCheck']   = self.path_remoteLocationCheck(d_msg)
            self.qprint(str(d_ret), comms = 'rx')
            if not d_ret['remoteCheck']['status']:
                self.qprint('An error occurred while checking the remote server.',
                            comms = 'error')
                d_ret['remoteCheck']['msg']     = "The remote path spec is invalid!"
                b_OK        = False
            else:
                d_ret['remoteCheck']['msg']     = "Check on remote path successful."
            d_transport['checkRemote']  = False
            d_ret['status']             = d_ret['localCheck']['status']
            d_ret['msg']                = d_ret['localCheck']['msg']

        b_jobExec           = False
        if b_OK:
            if 'compress' in d_transport and d_ret['status']:
                self.qprint('Calling %s_compress()...' % str_action, comms = 'status')
                d_ret['compress']   = eval("self.%s_compress(d_msg, **kwargs)" % str_action)
                d_ret['status']     = d_ret['compress']['status']
                d_ret['msg']        = d_ret['compress']['msg']
                b_jobExec       = True

            if 'copy' in d_transport:
                self.qprint('Calling %s_copy()...' % str_action, comms = 'status')
                d_ret['copyOp']     = eval("self.%s_copy(d_msg, **kwargs)" % str_action)
                d_ret['status']     = d_ret['copyOp']['copy']['status']
                d_ret['msg']        = d_ret['copyOp']['copy']['msg']
                b_jobExec       = True

        if not b_jobExec:
            d_ret['status']   = False
            d_ret['msg']      = 'No push/pull operation was performed! A filepath check failed!'

        if self.b_oneShot:
            d_ret['shutdown'] = self.server_ctlQuit(d_msg)
        return {'stdout': d_ret}

    def server_ctlQuit(self, d_msg):
        """

        :return: d_shutdown shut down JSON message from remote service.
        """

        d_shutdown      = {}
        d_meta          = d_msg['meta']
        d_meta['ctl'] = {
            'serverCmd': 'quit'
        }

        self.qprint('Attempting to shut down remote server...', comms = 'status')
        try:
            # d_shutdown  = self.push_core(d_msg, fileToPush = None)
            d_shutdown  = self.push_core(d_msg)
        except:
            pass
        return d_shutdown

    def pushPath(self, d_msg, **kwargs):
        """
        Push data to a remote server using pycurl.

        Essentially, this method is the central dispatching nexus to various
        specialized push operations.

        """

        return self.pathOp_do(d_msg, action = 'push')

    def pullPath(self, d_msg, **kwargs):
        """
        Pulls data from a remote server using pycurl.

        This method assumes that a prior call has "setup" a remote fileio
        listener and has the ip:port of that instance.

        Essentially, this method is the central dispatching nexus to various
        specialized pull operations.

        :param d_msg:
        :param kwargs:
        :return:
        """

        return self.pathOp_do(d_msg, action = 'pull')

    def httpStr_parse(self, **kwargs):

        for k,v in kwargs.items():
            if k == 'http':     self.str_http   = v

        # Split http string into IP:port and URL
        str_IPport          = self.str_http.split('/')[0]
        self.str_URL        = '/' + '/'.join(self.str_http.split('/')[1:])
        try:
            (self.str_ip, self.str_port) = str_IPport.split(':')
        except:
            self.str_ip     = str_IPport.split(':')
            self.str_port   = args.str_port

    def __call__(self, *args, **kwargs):
        """
        Main entry point for "calling".

        :param self:
        :param kwargs:
        :return:
        """
        str_action  = ''

        for key,val in kwargs.items():
            if key == 'msg':
                self.str_msg    = val
                self.d_msg      = json.loads(self.str_msg)
            if key == 'http':       self.httpStr_parse( http    = val)
            if key == 'verb':       self.str_verb               = val

        if len(self.str_msg):
            if 'action' in self.d_msg: str_action  = self.d_msg['action']
            if 'path' in str_action.lower():
                d_ret = self.pathOp_do(self.d_msg, action = str_action)
            else:
                if self.str_verb == 'GET':
                    d_ret = self.pull_core(msg = self.d_msg)
                if self.str_verb == 'POST':
                    d_ret = self.push_core(self.d_msg)
            str_stdout  = json.dumps(d_ret)
        else:
            d_ret = self.pull_core()
            str_stdout  = '%s' % d_ret

        if not self.b_quiet: print(Colors.CYAN)
        return(str_stdout)


#!/usr/bin/env python3.5

"""
    NAME

        C_snode, C_snodeBranch, C_stree

    DESCRIPTION

        These three classes are used to construct tree-structures
        composed of 'C_snode' instances. Branches contain dictionaries
        of C_snodes, while leaves contain dictionaries of a specific
        external data class.

    HISTORY
        o 26 March 2008
          Initial design and coding

        o 25 February 2014
          Resurrection as part of the fnndsc-netmap app.

        o 27 February 2014
          Cleanup, refactoring, re-indenting.
          Fix class/instance variable mixups.
"""

# System modules
import  os
import  sys
import  re
from    string                  import  *
import  pickle
import  json
import  collections
import  pudb

# pman local dependencies
from    .C_stringCore           import  *
from    .message                import  Message

class C_meta:
        """
        A "base" class containing 'meta' data pertinent to a node.
        """

        def __init__(self, al_mustInclude          = [],
                           al_mustNotInclude       = []):
            self._hitCount              = 0
            self.l_canInclude           = []
            self.l_mustInclude          = al_mustInclude
            self.l_mustNotInclude       = al_mustNotInclude
            self.b_printPre             = False
            self.sCore                  = C_stringCore()
            self._depth                 = 0
            self.str_pre                = ' '

        def __iter__(self):
            for key in ['_depth']:
                yield(key, getattr(self, key))

        #
        ## Getters/setters

        def pre(self, *args):
            """
            Get / set the str_pre
            """
            if len(args):
                self.str_pre = args[0]
            else:
                return self.str_pre

        def mustInclude(self, *args):
            """
            Get / set the [mustInclude].
            """
            if len(args):
                self.l_mustInclude = args[0]
            else:
                return l_mustInclude

        def mustNotInclude(self, *args):
            """
            Get / set the [mustInclude].
            """
            if len(args):
                self.l_mustNotInclude = args[0]
            else:
                return l_mustNotInclude

        def canInclude(self, *args):
            """
            Get / set the [mustInclude].
            """
            if len(args):
                self.l_canInclude = args[0]
            else:
                return l_canInclude

        def depth(self, *args):
            """
            Get/set the depth
            """
            if len(args):
                self._depth = args[0]
            else:
                return self._depth

        #
        ## core overloads

        def __str__(self):
            self.sCore.write('%s   +--depth............ %d\n' % (self.str_pre, self._depth))
            self.sCore.write('%s   +--hitCount......... %d\n' % (self.str_pre, self._hitCount))
            self.sCore.write('%s   +--mustInclude...... %s\n' % (self.str_pre, self.l_mustInclude))
            self.sCore.write('%s   +--mustNotInclude... %s\n' % (self.str_pre, self.l_mustNotInclude))
            str_ret = self.sCore.strget()
            self.sCore.reset()
            return str_ret

class C_snode:
        """
        A "container" node class. This container is the
        basic building block for larger tree-like database
        structures.

        The C_snode defines a single 'node' in this tree. It contains
        two lists, 'l_mustInclude' and 'l_mustNotInclude' that define
        the features described in the 'd_nodes' dictionary. This
        dictionary can in turn contain other C_snodes.
        """

        #
        # Methods
        #
        def __init__(self,      astr_nodeName           = "",
                                al_mustInclude          = [],
                                al_mustNotInclude       = []
                                ):
            #       - Core variables
            self.str_obj        = 'C_snode'      # name of object class
            self.str_name       = 'void'         # name of object variable
            self._id            = -1             # id of agent
            self._iter          = 0              # current iteration in an
                                                 #       arbitrary processing
                                                 #       scheme
            self._verbosity     = 0              # debug related value for
                                                 #       object
            self._warnings      = 0              # show warnings

            self.sCore          = C_stringCore()

            # The d_nodes is the basic building block of the C_snode container
            #+ class. It is simply a dictionary that contains 'nodes' that
            #+ satisfy a given feature set described by 'mustInclude' and
            #+ 'mustNotInclude'.
            #+
            #+ In general:
            #+  'snode_parent'      :       the parent node of this node -- useful
            #+                              for tree structuring.
            #+  '_hitCount'         :       count of hits for all items branching
            #+                              at this level. At the leaf level, this
            #+                              contains the length of 'contents'.
            #+  'l_mustInclude'     :       descriptive trait for specific feature
            #+                              level
            #+  'l_mustNotInclude'  :       exclusion trait for specific feature
            #+                              level
            #+  'd_nodes'           :       dictionary of child nodes branching
            #+                              from this node
            #+  'd_data'            :       a dictionary of data for *this* node
            #+
            #+ The pattern of 'mustInclude' and 'mustNotInclude' uniquely
            #+ characterizes a particular level. "Deeper" features (i.e. features
            #+ further along the dictionary tree) must satisfy the combined set
            #+ described by all the 'mustInclude' and 'mustNotInclude' traits of
            #+ each higher level.

            # self.meta                   = C_meta()
            self.snode_parent           = None
            self.d_nodes                = {}
            self.d_data                 = {}
            # self.b_printMetaData        = True
            self.b_printContents        = True
            self.b_printPre             = False
            self.str_nodeName           = astr_nodeName
            self.b_printPre             = False

        def __iter__(self):
            # yield('meta', dict(self.meta))
            if len(self.d_data):
                for key in self.d_data.keys():
                    yield(key, self.d_data[key])
            for key in self.d_nodes:
                yield(self.d_nodes[key].str_nodeName, dict(self.d_nodes[key]))

        #
        # Getters and setters

        def metaData_print(self, *args):
            return True
            # if len(args):
            #     self.b_printMetaData    = args[0]
            #     return True
            # else:
            #     return self.b_printMetaData

        def depth(self, *args):
            """
            Get/set the depth of this node.
            """
            if len(args):
                pass
                # self.meta.depth(args[0])
            else:
                # return self.meta.depth()
                return 0

        def printPre(self, *args):
            """
            get/set the str_pre string.
            """
            if len(args):
                self.b_printPre = args[0]
            else:
                return self.b_printPre

        @staticmethod
        def str_blockIndent(astr_buf, a_tabs=1, a_tabLength=4, **kwargs):
            """
            For the input string <astr_buf>, replace each '\n'
            with '\n<tab>' where the number of tabs is indicated
            by <a_tabs> and the length of the tab by <a_tabLength>

            Trailing '\n' are *not* replaced.
            """
            str_tabBoundary = " "
            for key, value in kwargs.items():
              if key == 'tabBoundary':  str_tabBoundary = value
            b_trailN = False
            length = len(astr_buf)
            ch_trailN = astr_buf[length - 1]
            if ch_trailN == '\n':
              b_trailN = True
              astr_buf = astr_buf[0:length - 1]
            str_ret = astr_buf
            str_tab = ''
            str_Indent = ''
            for i in range(a_tabLength):
                str_tab = '%s ' % str_tab
            str_tab = "%s%s" % (str_tab, str_tabBoundary)
            for i in range(a_tabs):
                str_Indent = '%s%s' % (str_Indent, str_tab)
            str_ret = re.sub('\n', '\n%s' % str_Indent, astr_buf)
            str_ret = '%s%s' % (str_Indent, str_ret)
            if b_trailN: str_ret = str_ret + '\n'
            return str_ret

        def __str__(self):
            self.sCore.reset()
            str_pre     = ""
            if not self.depth():
                str_pre = "o"
            else:
                str_pre = "+"
            self.sCore.write('%s---%s\n' % (str_pre, self.str_nodeName))
            if self.b_printPre:
                str_pre = "|"
            else:
                str_pre = " "
            # self.meta.pre(str_pre)
            # if self.b_printMetaData: self.sCore.write('%s' % self.meta)

            for key, value in self.d_data.items():
                self.sCore.write('%s   +--%-17s %s\n' % (str_pre, key, value))

            nodeCount     = len(self.d_nodes)
            if nodeCount and self.b_printContents:
                self.sCore.write('%s   +---+\n' % str_pre )
                elCount   = 0
                lastKey   = list(self.d_nodes)[-1]
                for node in list(self.d_nodes):
                    self.d_nodes[node].printPre(True)
                    if node == lastKey:
                        self.d_nodes[node].printPre(False)
                    str_contents = C_snode.str_blockIndent('%s' %
                        self.d_nodes[node], 1, 8, tabBoundary = "")
                    # str_contents = re.sub(r'                ', 'xxxxxxxx|xxxxxxx', str_contents)
                    if self.d_nodes[node].printPre():
                        str_contents = re.sub(r'                ', '        |       ', str_contents)
                    self.sCore.write(str_contents)
                    elCount   = elCount + 1
            return self.sCore.strget()

        #
        # Simple error handling
        def error_exit(self, astr_action, astr_error, astr_code):
            print("%s: FATAL error occurred"                % self.str_obj)
            print("While %s,"                               % astr_action)
            print("%s"                                      % astr_error)
            print("\nReturning to system with code %s\n"    % astr_code)
            sys.exit(astr_code)

        def node_branch(self, al_keys, al_values):
            """
            For each node in <al_values>, add to internal contents
            dictionary using key from <al_keys>.
            """
            if len(al_keys) != len(al_values):
                self.error_exit("adding branch nodes", "#keys != #values", 1)
            ldict = dict(zip(al_keys, al_values))
            self.node_dictBranch(ldict)

        def node_dictBranch(self, adict):
            """
            Expands the internal md_nodes with <adict>
            """
            self.d_nodes.update(adict)

class C_snodeBranch:
        """
        The C_snodeBranch class is basically a dictionary collection
        of C_snodes. Conceptually, a C_snodeBranch is a single "layer"
        of C_snodes all branching from a common ancestor node.
        """
        #
        # Member variables
        #


        #
        # Methods
        #

        def __str__(self):
            self.sCore.reset()
            for node in self.dict_branch.keys():
              self.sCore.write('%s' % self.dict_branch[node])
            return self.sCore.strget()

        def __init__(self, al_branchNodes):
            """
            Constructor.

            If instantiated with a list of nodes, will create/populate
            internal dictionary with appropriate C_snodes.
            """

            self.str_obj                = 'C_snodeBranch';  # name of object class
            self.str_name               = 'void';           # name of object variable
            self._id                    = -1;               # id of agent
            self._iter                  = 0;                # current iteration in an
                                                            #       arbitrary processing
                                                            #       scheme
            self._verbosity             = 0;                # debug related value for
                                                            #       object
            self._warnings              = 0;                # show warnings

            self.dict_branch            = {}
            self.sCore                  = C_stringCore()
            element                     = al_branchNodes[0]
            if isinstance(element, C_snode):
              for node in al_branchNodes:
                self.dict_branch[node]  = node
            else:
              for node in al_branchNodes:
                self.dict_branch[node]  = C_snode(node)

        #
        # Simple error handling
        def error_exit(self, astr_action, astr_error, astr_code):
            print("%s: FATAL error occurred"                % self.str_obj)
            print("While %s,"                               % astr_action)
            print("%s"                                      % astr_error)
            print("\nReturning to system with code %s\n"    % astr_code)
            sys.exit(astr_code)

        def node_branch(self, astr_node, abranch):
            """
            Adds a branch to a node, i.e. depth addition. The given
            node's md_nodes is set to the abranch's mdict_branch.
            """
            self.dict_branch[astr_node].node_dictBranch(abranch.dict_branch)

class C_stree:
        """
        The C_stree class provides methods for creating / navigating
        a tree composed of C_snodes.

        A C_stree is an ordered (and nested) collection of C_snodeBranch
        instances, with additional logic to match nodes with their parent
        node.

        The metaphor designed into the tree structure is that of a UNIX
        directory tree, with equivalent functions for 'cdnode', 'mknode'
        'lsnode'.
        """

        #
        # Methods
        #

        def metaData_print(self, *args):
            if len(args):
                # self.b_printMetaData    = args[0]
                return True
            else:
                # return self.b_printMetaData
                return True

        def log(self, *args):
            """
            get/set the log object.

            Caller can further manipulate the log object with object-specific
            calls.
            """
            if len(args):
                self._log = args[0]
            else:
                return self._log

        def name(self, *args):
            """
            get/set the descriptive name text of this object.
            """
            if len(args):
                self.__name = args[0]
            else:
                return self.__name

        def __init__(self, **kwargs):
            """
            Creates a tree structure and populates the "root"
            branch.
            """
            #
            # Member variables
            #
            #       - Core variables

            self.b_debugToFile          = True
            self.str_debugFile          = '/dev/null'
            for key,val in kwargs.items():
                if key == 'debugFile':      self.str_debugFile  = val
                if key == 'debugToFile':    self.b_debugToFile  = val

            self.debug                  = Message(logTo = self.str_debugFile)
            self.debug._b_syslog        = True
            self._log                   = Message()
            self._log._b_syslog         = True
            self.__name                 = "C_stree"

            self.str_obj                = 'C_stree'     # name of object class
            self.str_name               = 'void'        # name of object variable
            self._id                    = -1            # id of agent
            self._iter                  = 0             # current iteration in an
                                                        #       arbitrary processing
                                                        #       scheme
            self._verbosity             = 0             # debug related value for
                                                        #       object
            self._warnings              = 0             # show warnings
            self.b_printMetaData        = False

            self.l_allPaths             = []            # Each time a new C_snode is
                                                        #+ added to the tree, its path
                                                        #+ list is appended to this
                                                        #+ list variable.
            self.l_allFiles             = []            # A list of lists of of all files
                                                        #+ in the tree
            self.l_lwd                  = []            # A scratch path list variable
                                                        #+ for the lwd() method.
            self.l_fwd                  = []            # A scratch path list variable
                                                        #+ for the fwd() method.

            self.b_initFromDict         = False
            adict                       = {}
            al_rootBranch               = []
            for key,value in kwargs.items():
                if key == 'rootBranch': al_rootBranch   = value
                if key == 'dict':
                    self.b_initFromDict = True
                    adict               = value

            if not len(al_rootBranch):
                al_rootBranch           = ['/']
            if len(al_rootBranch):
                if not isinstance(al_rootBranch, list):
                    al_rootBranch       = ['/']
            self.sCore                  = C_stringCore()
            str_treeRoot                = '/'
            self.l_cwd                  = [str_treeRoot]
            self.sbranch_root           = C_snodeBranch([str_treeRoot])
            self.snode_current          = None
            self.snode_root             = self.sbranch_root.dict_branch[str_treeRoot]
            self.snode_root.depth(0)
            self.snode_root.snode_parent = self.snode_root
            self.root()
            self.l_allPaths             = self.l_cwd[:]
            if len(al_rootBranch) and al_rootBranch != ['/']:
                self.mknode(al_rootBranch)
            self.cd('/')
            if self.b_initFromDict:
                self.initFromDict(adict)

        @staticmethod
        def flatten(d, parent_key='', sep='/'):
            items = []
            for k, v in d.items():
                new_key = parent_key + sep + k if parent_key else k
                if isinstance(v, collections.MutableMapping):
                    if v == {}: v = {'contents': None}
                    items.extend(C_stree.flatten(v, new_key, sep=sep).items())
                else:
                    items.append((sep + new_key, v))
            return dict(items)

        def initFromDict(self, adict, **kwargs):
            """
            Initialize self from dictionary.
            :param adict:
            :return:
            """
            # First, flatten the dictionary into a dictionary of paths/files
            a_flat  = C_stree.flatten(adict)

            l_dir   = []
            # Now, build a tree from this structure by generating a list of paths
            for k, v in a_flat.items():
                l_dir.append(k.split('/')[1:-1])

            # remove duplicates...
            l_dir = [list(x) for x in set(tuple(x) for x in l_dir)]

            # build sorted list of paths...
            l_path  = ['/' + '/'.join(p) for p in l_dir]
            l_path.sort()

            # build the tree
            for dir in l_path: self.mkdir(dir)

            # and now add the leaves
            for file,contents in a_flat.items(): self.touch(file, contents)

        def __str__(self):
            self.sCore.reset()
            self.sCore.write('%s' % self.snode_root)
            return self.sCore.strget()

        def __iter__(self):
            yield(dict(self.snode_root))

        def root(self):
            """
            Reset all nodes and branches to 'root'.
            """
            str_treeRoot                = '/'
            self.l_cwd                  = [str_treeRoot]
            self.snode_current          = self.snode_root
            self.sbranch_current        = self.sbranch_root


        def cwd(self):
            """
            Return a UNIX FS type string of the current working 'directory'.
            """
            l_cwd                       = self.l_cwd[:]
            str_cwd                     = '/'.join(l_cwd)
            if len(str_cwd)>1: str_cwd  = str_cwd[1:]
            return str_cwd

        def path_has(self, **kwargs):
            """
            Checks if the current path has a node spec'd by kwargs
            """
            str_node    = "/" # This node will always be "False"
            for key, val in kwargs.items():
                if key == 'node':   str_node = val
            if str_node in self.l_cwd:
                return { 'found':   True,
                         'indices': [i for i, x in enumerate(self.l_cwd) if x == str_node]}
            else:
                return { 'found':   False,
                         'indices': [-1]}

        def pwd(self, **kwargs):
            """
            Returns the cwd

            Optional kwargs:

                node = <node>
                If specified, return only the directory name at depth <node>.

            """

            b_node  = False
            node    = 0
            for key,val in kwargs.items():
                if key == 'node':
                    b_node  = True
                    node    = int(val)

            str_path = self.cwd()
            if b_node:
                l_path      = str_path.split('/')
                if len(l_path) >= node+1:
                    str_path    = str_path.split('/')[node]
            return str_path

        def ptree(self):
            """
            Return all the paths in the tree.
            """
            return self.l_allPaths

        def node_mustNotInclude(self, al_mustNotInclude, ab_reset=False):
            """
            Either appends or resets the <mustNotInclude> list of snode_current
            depending on <ab_reset>.
            """
            if ab_reset:
                self.snode_current.l_mustNotInclude = al_mustNotInclude[:]
            else:
                l_current   = self.snode_current.l_mustNotInclude[:]
                l_total     = l_current + al_mustNotInclude
                self.snode_current.l_mustNotInclude = l_total[:]

        def node_mustInclude(self, al_mustInclude, ab_reset=False):
            """
            Either appends or resets the <mustInclude> list of snode_current
            depending on <ab_reset>.
            """
            if ab_reset:
                self.snode_current.l_mustInclude = al_mustInclude[:]
            else:
                l_current   = self.snode_current.l_mustInclude[:]
                l_total     = l_current + al_mustInclude
                self.snode_current.l_mustInclude = l_total[:]

        def paths_update(self, al_branchNodes):
            """
            Add each node in <al_branchNodes> to the self.ml_cwd and
            append the combined list to ml_allPaths. This method is
            typically not called by a user, but by other methods in
            this module.

            Returns the list of all paths.
            """
            for node in al_branchNodes:
                #print "appending %s" % node
                l_pwd       = self.l_cwd[:]
                l_pwd.append(node)
                #print "l_pwd: %s" % l_pwd
                #print "ml_cwd: %s" % self.ml_cwd
                self.l_allPaths.append(l_pwd)
            return self.l_allPaths

        def mkdir(self, astr_dirSpec):
            """
            Given an <astr_dirSpec> in form '/a/b/c/d/.../f',
            create that path in the internal stree, creating all
            intermediate nodes as necessary

            :param astr_dirSpec:
            :return:
            """
            if astr_dirSpec != '/' and astr_dirSpec != "//":
                str_currentPath = self.cwd()
                l_pathSpec = astr_dirSpec.split('/')
                if not len(l_pathSpec[0]):
                    self.cd('/')
                    l_nodesDepth = l_pathSpec[1:]
                else:
                    l_nodesDepth = l_pathSpec
                for d in l_nodesDepth:
                    self.mkcd(d)
                self.cd(str_currentPath)

        def mknode(self, al_branchNodes):
            """
            Create a set of nodes (branches) at current node. Analogous to
            a UNIX mkdir call, however nodes can be any type (i.e. not
            just "directories" but also "files")
            """
            b_ret = True
            # First check that none of these nodes already exist in the tree
            l_branchNodes = []
            for node in al_branchNodes:
                l_path      = self.l_cwd[:]
                l_path.append(node)
                #print l_path
                #print self.ml_allPaths
                #print self.b_pathOK(l_path)
                if not self.b_pathOK(l_path):
                    l_branchNodes.append(node)
            if not len(l_branchNodes):
                return False
            snodeBranch   = C_snodeBranch(l_branchNodes)
            for node in l_branchNodes:
                depth = self.snode_current.depth()
                # if (self.msnode_current != self.msnode_root):
                snodeBranch.dict_branch[node].depth(depth+1)
                snodeBranch.dict_branch[node].snode_parent = self.snode_current
            self.snode_current.node_dictBranch(snodeBranch.dict_branch)
            # Update the ml_allPaths
            self.paths_update(al_branchNodes)
            return b_ret

        def mkcd(self, astr_node):
            """Creates a single node and cd's into that node
            """
            self.mknode([astr_node])
            return self.cdnode(astr_node)

        def isdir(self, str_path):
            """
            A convenience function, returns bool if <str_path> is
            a "dir" in the tree space
            """

            return self.b_pathInTree(str_path)[0]

        def isfile(self, str_path):
            """
            A convenience function, returns bool if <str_path> is
            a "file" in the tree space
            """

            b_isFile            = False
            if self.isdir(str_path):
                b_isFile        = False
            else:
                str_parentDir   = '/'.join(str_path.split('/')[0:-1])
                str_fileName    = str_path.split('/')[-1]
                if self.cd(str_parentDir)['status']:
                    l_files = self.lsf(str_parentDir)
                    if any(str_fileName in f for f in l_files):
                        b_isFile    = True
            return b_isFile

        def exists(self, fileDirSpec, **kwargs):
            """
            Simple returns a boolean if the <fileDirSpec>
            exists in the current dir (or dir spec'd by **kwargs)
            """

            str_path        = self.cwd()
            for k,v in kwargs.items():
                if k == 'path': str_path    = v
            if fileDirSpec in self.lstr_lsnode(str_path):
                return True
            if fileDirSpec in self.lsf(str_path):
                return True
            return False

        def cat(self, name):
            """
            Returns the contents of the 'name'd element at this level.

            If file does not exist, returns a False

            TODO: parse possible path spec in name...
            """

            origDir = self.cwd()
            # First, parse any path specs...
            ret     = None
            path    = '/'.join(name.split('/')[0:-1])
            if len(path):
                name    = name.split('/')[-1]
                ret     = self.cd(path)

            if name in self.snode_current.d_data:
                ret     =  self.snode_current.d_data[name]
            else:
                ret     = False
            self.cd(origDir)
            return ret

        def graft(self, atree, apath = '/'):
            """

            NB: THIS METHOD HAS UNEXPECTED MEMORY BEHAVIOUR!
                USE THE copy() METHOD INSTEAD!

            Attach (link) apath in a separate atree to here.

            Functionally equivalent to

                ln -s atree:/apath .

            Also updates the self tree's path list space.

            :param atree: a tree
            :param apath: a path in atree
            :return: True | False depending on successful graft
            """
            ret = False
            if atree.cd(apath)['status']:
                atree_nodes                 = atree.lstr_lsnode()
                str_self_cwd                = self.cwd()
                str_atree_cwd               = atree.cwd()

                b_childrenLink              = False
                if apath != '/':
                    l_apath = apath.split('/')
                    # No trailing '/'
                    if l_apath[-1] != '':
                        self.mkcd(apath.split('/')[-1])
                        self.snode_current.d_nodes  = atree.snode_current.d_nodes
                        self.snode_current.d_data   = atree.snode_current.d_data
                    # Trailing '/'
                    else:
                        apath = apath[0:-1]
                        if apath[-1] != '/':
                            atree.cd(apath)
                            atree_nodes     = atree.lstr_lsnode()
                            b_childrenLink  = True
                else:
                    b_childrenLink  = True
                if b_childrenLink:
                    self.mknode(atree_nodes)
                    for node in atree_nodes:
                        self.cd(node)
                        atree.cd(apath + '/' + node)
                        self.snode_current.d_nodes  = atree.snode_current.d_nodes
                        self.snode_current.d_data   = atree.snode_current.d_data
                        atree.cd('../')
                        self.cd('../')

                # Update internal space of possible paths.
                self.cd('../')
                self.pathFromHere_explore(self.cwd())
                self.cd(str_self_cwd)
                atree.cd(str_atree_cwd)
                ret = True
            return ret

        def touch(self, name, data):
            """
            Create a 'file' analog called 'name' and put 'data' to the d_data dictionary
            under key 'name'.

            The 'name' can contain a path specifier.
            """
            b_OK        = True
            str_here    = self.cwd()
            # print("here!")
            # print(self.snode_current)
            # print(self.snode_current.d_nodes)
            l_path = name.split('/')
            if len(l_path) > 1:
                self.cd('/'.join(l_path[0:-1]))
            name = l_path[-1]
            self.snode_current.d_data[name] = data
            # print(self.snode_current)
            self.cd(str_here)
            return b_OK

        def rm(self, name):
            """
            Remove a data analog called 'name'.

            The 'name' can contain a path specifier.

            Warning: see

                http://stackoverflow.com/questions/5844672/delete-an-element-from-a-dictionary

            deleting from the snode_current changes dictionary contents for any other
            agents that have references to the same instance.

            This deletes either directories or files.

            """
            b_OK        = False
            str_here    = self.cwd()
            l_path = name.split('/')
            if len(l_path) > 1:
                self.cd('/'.join(l_path[0:-1]))
            name = l_path[-1]
            if name in self.snode_current.d_data:
                del self.snode_current.d_data[name]
                b_OK    = True
            if name in self.snode_current.d_nodes:
                del self.snode_current.d_nodes[name]
                b_OK    = True
            self.cd(str_here)
            return b_OK

        def append(self, name, data):
            """Append 'data' to the current node d_data

            This method appends 'data' to the current contents in the
            key named 'name'. The append assumes that the operation
            makes sense and that the data types can be appended to
            each other.

            """
            b_OK = True
            self.snode_current.d_data[name] = self.snode_current.d_data[name] + data
            return b_OK


        def b_pathOK(self, al_path):
            """
            Checks if the absolute path specified in the al_path
            is valid for current tree
            """
            b_OK  = True
            try:          self.l_allPaths.index(al_path)
            except:       b_OK    = False
            return b_OK

        def b_pathInTree(self, astr_path):
            """
            Converts a string <astr_path> specifier to a list-based
            *absolute* lookup, i.e. "/node1/node2/node3" is converted
            to ['/' 'node1' 'node2' 'node3'].

            The method also understands a paths that start with: '..' or
            combination of '../../..' and is also aware that the root
            node is its own parent.

            If the path list conversion is valid (i.e. exists in the
            space of existing paths, l_allPaths), return True and the
            destination path list; else return False and the current
            path list.
            """
            if astr_path == '/':  return True, ['/']
            al_path               = astr_path.split('/')
            # Do we have a trailing '/' and not doing a '../'? If so, strip it..!
            if astr_path != '../' and al_path[-1] == '':
                al_path = al_path[0:-2]

            # Check for absolute path
            if not len(al_path[0]):
                al_path[0]          = '/'
                # print "returning %s : %s" % (self.b_pathOK(al_path), al_path)
                return self.b_pathOK(al_path), al_path
            # Here we are in relative mode...
            # First, resolve any leading '..'
            l_path        = self.l_cwd[:]
            if al_path[0] == '..':
                while al_path[0] == '..' and len(al_path):
                    l_path    = l_path[0:-1]
                    if len(al_path) >= 2: al_path   = al_path[1:]
                    else: al_path[0] = ''
                    # print "l_path  = %s" % l_path
                    # print "al_path = %s (%d)" % (al_path, len(al_path[0]))
                if len(al_path[0]):
                    # print "extending %s with %s" % (l_path, al_path)
                    l_path.extend(al_path)
            else:
                l_path      = self.l_cwd
                l_path.extend(al_path)
            # print "final path list = %s (%d)" % (l_path, len(l_path))
            if len(l_path)>=1 and l_path[0] != '/':      l_path.insert(0, '/')
            if len(l_path)>1:            l_path[0]       = ''
            if not len(l_path):          l_path          = ['/']
            str_path      = '/'.join(l_path)
            # print "final path str  = %s" % str_path
            b_valid, al_path = self.b_pathInTree(str_path)
            return b_valid, al_path

        def cdnode(self, astr_path):
            """Change working node to astr_path.

            The path is converted to a list, split on '/'. By performing a 'cd'
            all parent and derived nodes need to be updated relative to
            new location.

            Args:
                astr_path (string): The path to cd to.

            Returns:
                {"status" : True/False , "path": l_cwd -- the path as list}

            """

            # Start at the root and then navigate to the
            # relevant node
            l_absPath             = []
            b_valid, l_absPath    = self.b_pathInTree(astr_path)
            if b_valid:
                #print "got cdpath = %s" % l_absPath
                self.l_cwd              = l_absPath[:]
                self.snode_current      = self.snode_root
                self.sbranch_current    = self.sbranch_root
                #print l_absPath
                for node in l_absPath[1:]:
                    self.snode_current = self.snode_current.d_nodes[node]
                self.sbranch_current.dict_branch = self.snode_current.snode_parent.d_nodes
                return {"status": True, "path": self.l_cwd}
            return {"status": False, "path": []}

        def cd(self, astr_path):
            """Alias for cdnode()

            """
            return self.cdnode(astr_path)

        def ls(self, astr_path="", **kwargs):
            b_lsData    = True
            b_lsNodes   = True
            str_cwd       = self.cwd()
            if len(astr_path): self.cdnode(astr_path)
            str_nodes   = self.str_lsnode(astr_path)
            d_data      = self.snode_current.d_data
            for key, val in kwargs.items():
                if key == 'data':   b_lsData    = val
                if key == 'nodes':  b_lsNodes   = val
            if len(astr_path): self.cdnode(str_cwd)
            if b_lsData and b_lsNodes:
                return str_nodes, d_data
            if b_lsData:
                return d_data
            if b_lsNodes:
                return str_nodes
            return str_nodes, d_data

        def lsf(self, astr_path=""):
            """
            List only the "files" in the astr_path.

            :param astr_path: path to list
            :return: "files" in astr_path, empty list if no files
            """
            d_files = self.ls(astr_path, nodes=False, data=True)
            l_files = d_files.keys()
            return l_files

        def str_lsnode(self, astr_path=""):
            """
            Print/return the set of nodes branching from current node as string
            """
            self.sCore.reset()
            str_cwd       = self.cwd()
            if len(astr_path): self.cdnode(astr_path)
            for node in self.snode_current.d_nodes.keys():
                self.sCore.write('%s\n' % node)
            str_ls = self.sCore.strget()
            if len(astr_path): self.cdnode(str_cwd)
            return str_ls

        def lstr_lsnode(self, astr_path=""):
            """
            Return the string names of the set of nodes branching from
            current node as list of strings
            """
            self.sCore.reset()
            str_cwd       = self.cwd()
            if len(astr_path): self.cdnode(astr_path)
            lst = self.snode_current.d_nodes.keys()
            if len(astr_path): self.cdnode(str_cwd)
            return lst

        def lsbranch(self, astr_path=""):
            """
            Print/return the set of nodes in current branch
            """
            self.sCore.reset()
            str_cwd       = self.cwd()
            if len(astr_path): self.cdnode(astr_path)
            self.sCore.write('%s' % self.sbranch_current.dict_branch.keys())
            str_ls = self.sCore.strget()
            if len(astr_path): self.cdnode(str_cwd)
            return str_ls

        def lstree(self, astr_path=""):
            """
            Print/return the tree from the current node.
            """
            self.sCore.reset()
            str_cwd       = self.cwd()
            if len(astr_path): self.cdnode(astr_path)
            str_ls        = '%s' % self.snode_current
            print(str_ls)
            if len(astr_path): self.cdnode(str_cwd)
            return str_ls

        def lsmeta(self, astr_path=""):
            """
            Print/return the "meta" information of the node, i.e.
                o mustInclude
                o mustNotInclude
                o hitCount
            """
            self.sCore.reset()
            str_cwd       = self.cwd()
            if len(astr_path): self.cdnode(astr_path)
            b_contentsFlag        = self.snode_current.b_printContents
            self.snode_current.b_printContents = False
            str_ls        = '%s' % self.snode_current
            print(str_ls)
            if len(astr_path): self.cdnode(str_cwd)
            self.snode_current.b_printContents  = b_contentsFlag
            return str_ls

        def tree_metaData_print(self, aval):
            self.metaData_print(aval)
            self.treeRecurse(self.treeNode_metaSet)

        def treeNode_metaSet(self, astr_path, **kwargs):
            """
            Sets the metaData_print bit on node at <astr_path>.
            """
            self.cdnode(astr_path)
            self.snode_current.metaData_print(self.b_printMetaData)
            return {'status': True}

        def node_save(self, astr_pathInTree, **kwargs):
            """
            Typically called by the explore()/recurse() methods and of form:

                f(pathInTree, **kwargs)

            and returns dictionary of which one element is

                'status':   True|False

            recursion continuation flag is returned:

                'continue': True|False

            to signal calling parent whether or not to continue with tree
            transversal.

            Save the node specified by a path in the data tree to disk.

            Given a "root" on the disk storage, create the path relative
            to that root, and in that location, save the contents of the internal
            node's d_data at that tree path location.

            :param kwargs:
            :return:
            """
            str_pathDiskRoot    = '/tmp'
            str_pathDiskOrig    = os.getcwd()
            srt_pathDiskFull    = ''
            str_pathTree        = ''
            str_pathTreeOrig    = self.pwd()
            b_failOnDirExist    = True
            b_saveJSON          = True
            b_savePickle        = False
            for key, val in kwargs.items():
                if key == 'startPath':      str_pathTree        = val
                if key == 'pathDiskRoot':   str_pathDiskRoot    = val
                if key == 'failOnDirExist': b_failOnDirExist    = val
                if key == 'saveJSON':       b_saveJSON          = val
                if key == 'savePickle':     b_savePickle        = val

            str_pathDiskFull    = str_pathDiskRoot + str_pathTree
            # print('\n')
            # print('In self.node_save():')
            # print('memTree:  %s' % (str_pathTree))
            # print('diskRoot: %s' % str_pathDiskRoot)
            # print('distPath: %s' % str_pathDiskFull)
            # print(kwargs.keys())
            # print(kwargs.values())
            # print('\n')

            if len(str_pathDiskRoot):
                if not os.path.isdir(str_pathDiskRoot):
                    # print('Processing path: %s' % str_pathDiskRoot)
                    try:
                        # print('mkdir %s' % str_pathDiskRoot)
                        os.makedirs(str_pathDiskRoot)
                    except OSError as exception:
                        return {'status' :      False,
                                'continue':     False,
                                'message':      'unable to create pathDiskRoot: %s' % str_pathDiskRoot,
                                'exception':    exception}
                # print('cd to %s' % str_pathDiskRoot)
                os.chdir(str_pathDiskRoot)
                if self.cd(str_pathTree)['status']:
                    if str_pathTree != '/':
                        # print('mkdir %s' % str_pathDiskFull)
                        try:
                            os.makedirs(str_pathDiskFull)
                        except OSError as exception:
                            if b_failOnDirExist:
                                return {'status' :      False,
                                        'continue':     False,
                                        'message':      'unable to create pathDiskRoot: %s' % str_pathDiskRoot,
                                        'exception':    exception}

                    os.chdir(str_pathDiskFull)
                    for str_filename, contents in self.snode_current.d_data.items():
                        # print("str_filename = %s; contents = %s" % (str_filename, contents))
                        if b_saveJSON:
                            with open(str_filename, 'w')    as f: json.dump(contents,     f)
                            f.close()
                        if b_savePickle:
                            with open(str_filename, 'wb')   as f: json.dump(contents,     f)
                            f.close()
                else:
                    return{'status':    False,
                           'continue':  False,
                           'message':   'pathTree invalid'}
                self.cd(str_pathTreeOrig)
                os.chdir(str_pathDiskOrig)
                return {'status':   True,
                        'continue': True}
            return {'status':   False,
                    'continue': False,
                    'message':  'pathDisk not specified'}

        def node_copy(self, astr_pathInTree, **kwargs):
            """
            Typically called by the explore()/recurse() methods and of form:

                f(pathInTree, **kwargs)

            and returns dictionary of which one element is

                'status':   True|False

            recursion continuation flag is returned:

                'continue': True|False

            to signal calling parent whether or not to continue with tree
            transversal.

            Save the node specified by a path in the data tree of self
            (the astr_pathInTree) to the passed data tree, relative to a
            passed 'pathDiskRoot', i.e.

                S.node_copy('/', destination = T, pathDiskRoot = '/some/path/in/T')

            Will copy the items and "folders" in (source) S:/ to
            (target) T:/some/path/in/T

            :param kwargs:
            :return:
            """

            # Here, 'T' is the target 'disk'.
            T                   = None
            str_pathDiskRoot    = ''
            str_pathDiskFull    = ''
            str_pathTree        = ''
            str_pathTreeOrig    = self.pwd()
            for key, val in kwargs.items():
                if key == 'startPath':      str_pathTree        = val
                if key == 'pathDiskRoot':   str_pathDiskRoot    = val
                if key == 'destination':    T                   = val
            str_pathDiskOrig    = T.pwd()
            str_pathDiskFull    = str_pathDiskRoot + str_pathTree

            # self.debug('In node_copy... str_pathDiskfull = %s\n' % str_pathDiskFull)

            if len(str_pathDiskFull):
                if not T.isdir(str_pathDiskFull):
                    try:
                        T.mkdir(str_pathDiskFull)
                    except:
                        return {'status' :      False,
                                'continue':     False,
                                'message':      'unable to create pathDiskFull: %s' % str_pathDiskFull,
                                'exception':    exception}
                if T.cd(str_pathDiskFull)['status']:
                    if self.cd(str_pathTree)['status']:
                        T.cd(str_pathDiskFull)
                        for str_filename, contents in self.snode_current.d_data.items():
                            # print("str_filename = %s; contents = %s" % (str_filename, contents))
                            T.touch(str_filename, contents)
                    else:
                        return{'status':    False,
                               'continue':  False,
                               'message':   'source pathTree invalid'}
                    self.cd(str_pathTreeOrig)
                    T.cd(str_pathDiskOrig)
                    return {'status':   True,
                            'continue': True}
                else:
                    return{'status':    False,
                           'continue':  False,
                           'message':   'target pathDiskFull invalid'}
            return {'status':   False,
                    'continue': False,
                    'message':  'pathDiskFull not specified'}

        def copy(self, **kwargs):
            """

            Convenience function/alias for tree_copy()

            :param kwargs:
            :return:
            """
            return self.tree_copy(**kwargs)

        def tree_copy(self, **kwargs):
            """
            Deep copy "self" to a target <destination = T>.

            Essentially, this creates/copies this self tree to
            a target tree.

            For kwargs, see node_save()

            :param kwargs:
            :return:
            """
            # pudb.set_trace()
            kwargs['f']         = self.node_copy
            return self.treeExplore(**kwargs)

        def tree_save(self, **kwargs):
            """
            Save a tree to disk.

            Essentially, this creates a mirror of the internal tree
            structure on disk.

            For kwargs, see node_save()

            :param kwargs:
            :return:
            """

            kwargs['f']         = self.node_save
            self.treeExplore(**kwargs)

        @staticmethod
        def tree_load(**kwargs):
            """
            Load a tree from disk.

            Essentially, this reads a disk filetree into an snode tree.

            :param kwargs:
            :return:
            """
            str_pathDiskRoot    = ''
            b_loadJSON          = True
            b_loadPickle        = False
            for key, val in kwargs.items():
                if key == 'pathDiskRoot':   str_pathDiskRoot    = val
                if key == 'loadJSON':       b_loadJSON          = val
                if key == 'loadPickle':     b_loadPickle        = val

            l_dir   = []
            l_file  = []
            for str_root, l_directories, l_filenames in os.walk(str_pathDiskRoot):
                for str_dir in l_directories:
                    l_dir.append(os.path.join(str_root, str_dir))
                for str_file in l_filenames:
                    l_file.append(os.path.join(str_root, str_file))
            stree_dirs  = [d.replace(str_pathDiskRoot, '') for d in l_dir]
            stree_files = [f.replace(str_pathDiskRoot, '') for f in l_file]

            # Create the tree
            rtree       = C_stree()
            # Build the directory structures
            for d in stree_dirs:
                # print(d)
                rtree.mkdir(d)
            # Now read any files
            for f in stree_files:
                dirname     = os.path.dirname(f)
                filename    = os.path.basename(f)
                # print(f)
                if b_loadJSON:
                    with open(str_pathDiskRoot + '/' + f, 'r') as fp:
                        contents = json.load(fp)
                        fp.close()
                if b_loadPickle:
                    with open(str_pathDiskRoot + '/' + f, 'rb') as fp:
                        contents = pickle.load(fp)
                        fp.close()

                if rtree.cd(dirname)['status']:
                    rtree.touch(filename, contents)

            return rtree


        def treeExplore(self, **kwargs):
            """
            Recursively walk through a C_stree, applying a passed function
            at each node. The actual "walk" uses individual nodes' internal
            child dictionaries.

            It is assumed that the start origin of exploration can in fact
            be reached by a 'cd()' call. Directories are added to the internal
            l_allPaths list variable as they are discovered.

            kwargs:
                startPath=<startPath> :     The starting point in the tree
                func=<f> :                  The function to apply at each node

                Additional kwargs are passed to <f>

            <f> is a function that is called on a node path. It is of form:

                f(path, **kwargs)

            where path is a node in the tree space.

            <f> must return a dictionary containing at least one field:

                { "status": True | False }

            This same dictionary is also returned out to the caller of this
            function.

            """

            str_recursePath = ''
            str_startPath   = '/'
            f               = None
            ret             = {}

            for key,val in kwargs.items():
                if key == 'startPath':  str_startPath   = val
                if key == 'f':          f               = val

            # print 'processing node: %s' % str_startPath
            if self.cd(str_startPath)['status']:
                ret = f(str_startPath, **kwargs)
                if ret['status']:
                    for node in self.lstr_lsnode(str_startPath):
                        if str_startPath == '/': str_recursePath = "/%s" % node
                        else: str_recursePath = '%s/%s' % (str_startPath, node)
                        l_recursePath       = str_recursePath.split('/')
                        l_recursePath[0]    = '/'
                        if not l_recursePath in self.l_allPaths:
                            self.l_allPaths.append(l_recursePath)
                        kwargs['startPath'] = str_recursePath
                        self.treeExplore(**kwargs)
            else: ret['status']  = False
            return ret

        def treeWalk(self, **kwargs):
            """
            Recursively walk through a C_stree, applying a passed function
            at each node. The actual "walk" depends on using the 'cd' function
            which will only descend into paths that already exist in the
            internal path database.

            To flush this and explore the pathspace de novo, use treeExplore()
            instead.

            kwargs:
                --startPath=<startPath> :   The starting point in the tree
                --func=<f> :                The function to apply at each node

                Additional kwargs are passed to <f>

            <f> is a function that is called on a node path. It is of form:

                f(path, **kwargs)

            where path is a node in the tree space.

            <f> must return a dictionary containing at least one field:

                { "status": True | False }

            This same dictionary is also returned out to the caller of this
            function.

            """

            str_recursePath = ''
            str_startPath   = '/'
            f               = None
            ret             = {}

            for key,val in kwargs.items():
                if key == 'startPath':  str_startPath   = val
                if key == 'f':          f               = val

            # print 'processing node: %s' % str_startPath
            if self.cd(str_startPath)['status']:
                ret = f(str_startPath, **kwargs)
                if ret['status']:
                    for node in self.lstr_lsnode(str_startPath):
                        if str_startPath == '/': str_recursePath = "/%s" % node
                        else: str_recursePath = '%s/%s' % (str_startPath, node)
                        self.treeWalk(f = f, startPath = str_recursePath)
            return ret

        def treeRecurse(self, afunc_nodeEval = None, astr_startPath = '/'):
            """
            Recursively walk through the C_stree, starting from node
            <astr_startPath> and using the internal l_allPaths space
            as verifier.

            To self-discover tree structure based on internal node links,
            use treeExplore() instead.

            The <afunc_nodeEval> is a function that is called on a node
            path. It is of form:

                afunc_nodeEval(astr_startPath, **kwargs)

            and must return a dictionary containing at least one field:

                { "status": True | False }

            This same dictionary is also returned out to the caller of this
            function.

            """
            ret = {'status': False}
            [b_valid, l_path ] = self.b_pathInTree(astr_startPath)
            if b_valid and afunc_nodeEval:
                ret = afunc_nodeEval(astr_startPath)
            b_OK = ret['status']
            #print 'processing node: %s' % astr_startPath
            if b_OK:
                for node in self.lstr_lsnode(astr_startPath):
                    if astr_startPath == '/': recursePath = "/%s" % node
                    else: recursePath = '%s/%s' % (astr_startPath, node)
                    self.treeRecurse(afunc_nodeEval, recursePath)
            return ret

        def lwd(self, astr_startPath, **kwargs):
            """
            Return the cwd in treeRecurse compatible format.
            :return: Return the cwd in treeRecurse compatible format.
            """
            if self.cd(astr_startPath)['status']:
                self.l_lwd.append(self.cwd())

            return {'status': True, 'cwd': self.cwd()}

        def fwd(self, astr_startPath, **kwargs):
            """
            Return the files-in-working-directory in treeRecurse
            compatible format.
            :return: Return the cwd in treeRecurse compatible format.
            """
            status = self.cd(astr_startPath)['status']
            if status:
                l = self.lsf()
                if len(l):
                    lf = [self.cwd() + '/' + f for f in l]
                    for entry in lf:
                        self.l_fwd.append(entry)

            return {'status': status, 'cwd': self.cwd()}

        def filesFromHere_explore(self, astr_startPath = '/'):
            """
            Return a list of path/files from "here" in the stree, using
            the child explore access.

            :param astr_startPath: path from which to start
            :return:
            """
            self.l_fwd  = []
            self.treeExplore(startPath = astr_startPath, f=self.fwd)
            self.l_allFiles = [f.split('/') for f in self.l_fwd]
            for i in range(0, len(self.l_allFiles)):
                self.l_allFiles[i][0] = '/'
            return self.l_fwd

        def pathFromHere_walk(self, astr_startPath = '/'):
            """
            Return a list of paths from "here" in the stree, using
            the internal cd() to walk the path space.

            :return: a list of paths from "here"
            """

            self.l_lwd  = []
            self.treeWalk(startPath = astr_startPath, f=self.lwd)
            return self.l_lwd

        def pathFromHere_explore(self, astr_startPath = '/'):
            """
            Return a list of paths from "here" in the stree, using the
            child explore access.

            :param astr_startPath: path from which to start
            :return: a list of paths from "here"
            """

            self.l_lwd  = []
            self.treeExplore(startPath = astr_startPath, f=self.lwd)
            return self.l_lwd

        #
        # Simple error handling
        def error_exit(self, astr_action, astr_error, astr_code):
            print("%s: FATAL error occurred" % self.str_obj)
            print("While %s," % astr_action)
            print("%s" % astr_error)
            print("\nReturning to system with code %s\n" % astr_code)
            sys.exit(astr_code)

if __name__ == "__main__":


    aTree = C_stree()
    bTree = C_stree()
    ATree = C_stree()

    aTree.cd('/')
    aTree.mkcd('a')
    aTree.mknode(['b', 'c'])
    aTree.cd('b')
    aTree.touch('file1', 10)
    aTree.touch('file2', "Rudolph Pienaar")
    aTree.touch('file3', ['this', 'is', 'a', 'list'])
    aTree.touch('file4', ('this', 'is', 'a', 'tuple'))
    aTree.touch('file5', {'name': 'rudolph', 'address': '505 Washington'})

    aTree.mknode(['d', 'e'])
    aTree.cd('d')
    aTree.mknode(['h', 'i'])
    aTree.cd('/a/b/e')
    aTree.mknode(['j', 'k'])
    aTree.cd('/a/c')
    aTree.mknode(['f', 'g'])
    aTree.cd('f')
    aTree.mknode(['l', 'm'])
    aTree.cd('/a/c/g')
    aTree.mknode(['n', 'o'])

    ATree.cd('/')
    ATree.mkcd('A')
    ATree.mknode(['B', 'C'])
    ATree.cd('B')
    ATree.mknode(['D', 'E'])
    ATree.cd('D')
    ATree.mknode(['H', 'I'])
    ATree.cd('/A/B/E')
    ATree.mknode(['J', 'K'])
    ATree.cd('/A/B/E/K')
    ATree.touch('file1', 11)
    ATree.touch('file2', "Reza Pienaar")
    ATree.touch('file3', ['this', 'is', 'another', 'list'])
    ATree.touch('file4', ('this', 'is', 'another', 'tuple'))
    ATree.touch('file5', {'name': 'reza', 'address': '505 Washington'})
    ATree.cd('/A/C')
    ATree.mknode(['F', 'G'])
    ATree.cd('F')
    ATree.mknode(['L', 'M'])
    ATree.cd('/A/C/G')
    ATree.mknode(['N', 'O'])

    bTree.cd('/')
    bTree.mkcd('1')
    bTree.mknode(['2', '3'])
    bTree.cd('2')
    bTree.mknode(['4', '5'])
    bTree.cd('4')
    bTree.mknode(['8', '9'])
    bTree.cd('/1/2/5')
    bTree.mknode(['10', '11'])
    bTree.cd('/1/3')
    bTree.mknode(['6', '7'])
    bTree.cd('6')
    bTree.mknode(['12', '13'])
    bTree.cd('/1/3/7')
    bTree.mknode(['14', '15'])

    aTree.tree_metaData_print(False)
    ATree.tree_metaData_print(False)
    bTree.tree_metaData_print(False)

    print('aTree = %s' % aTree)
    # print(aTree.pathFromHere_walk('/'))
    print('ATree = %s' % ATree)
    # print(ATree.pathFromHere_walk('/'))
    print('bTree = %s' % bTree)
    # print(bTree.pathFromHere_walk('/'))

    aTree.cd('/')
    aTree.graft(bTree, '/1/2/')
    aTree.tree_metaData_print(False)
    print('aTree = %s' % aTree)
    # print(aTree.pathFromHere_walk('/'))
    # print(aTree.l_allPaths)

    bTree.cd('/1/2/4/9')
    bTree.graft(ATree, '/A/B')
    bTree.tree_metaData_print(False)
    print('bTree = %s' % bTree)
    # print(bTree.pathFromHere_walk('/'))
    # print(bTree.l_allPaths)

    print('aTree = %s' % aTree)
    # print(aTree.pathFromHere_explore('/'))
    # print(aTree.l_allPaths)
    # print(aTree.filesFromHere_explore('/'))
    # print(aTree.l_allFiles)

    print('Saving bTree...')
    bTree.tree_save(startPath       = '/',
                    pathDiskRoot    = '/tmp/bTree',
                    failOnDirExist  = True,
                    saveJSON        = True,
                    savePickle      = False)

    print('Saving aTree...')
    aTree.tree_save(startPath       = '/',
                    pathDiskRoot    = '/tmp/aTree',
                    failOnDirExist  = True,
                    saveJSON        = True,
                    savePickle      = False)

    print('Reading aTree into cTree...')
    cTree = C_stree.tree_load(
                    pathDiskRoot    = '/tmp/aTree',
                    loadJSON        = True,
                    loadPickle      = False)
    cTree.tree_metaData_print(False)
    print('cTree = %s' % cTree)
    cTree.rm('/4/9/B/E/K/file1')
    print('cTree = %s' % cTree)
    cTree.rm('/4/9/B/E/K/file2')
    print('cTree = %s' % cTree)
    cTree.rm('/4/9/B/E/K')
    print('cTree = %s' % cTree)

    dTree   = C_stree()
    cTree.tree_copy(startPath   = '/a/b/file5',
                    destination = dTree)
    print('dTree = %s' % dTree)



#!/usr/bin/env python3.5
# -*- coding: utf-8 -*-

import  abc
import  time
import  os

import  threading
import  zmq
from    webob           import  Response
import  psutil

import  queue
from    functools       import  partial
import  platform

import  multiprocessing

import  pudb
import  json


# pman local dependencies
from   ._colors           import Colors
from   .crunner           import crunner
from   .C_snode           import *
from   .debug             import debug
from   .pfioh             import *

import  docker

str_devNotes = """

    10 May 2017
    *   Should methods in the listener be functors? Certain methods, such as 'run' and 'status' need
        specialized implementations based on a run environment. This run environment is not known 
        by the listener when it starts, but can be specified at payload parsing by the process() 
        method. This, a method such as
        
            t_run_process()
            
        might need at arbitrary call time to be specialized to some external condition set (say by
        running as a container). Naively, this can be parsed in the message and thread redirected to
        
            t_run_process_swarm()
            
        for example.
        
        Would a functor type approach be useful at all?

"""


class StoppableThread(threading.Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self, *args, **kwargs):
        super(StoppableThread, self).__init__(*args, **kwargs)
        self._stopper = threading.Event()

    def stopit(self):
        self._stopper.set()

    def stopped(self):
        return self._stopper.isSet()

class pman(object):
    """
    The server class for the pman (process manager) server

    """
    __metaclass__   = abc.ABCMeta

    def col2_print(self, str_left, str_right):
        self.dp.qprint(Colors.WHITE +
              ('%*s' % (self.LC, str_left)), end='')
        self.dp.qprint(Colors.LIGHT_BLUE +
              ('%*s' % (self.RC, str_right)) + Colors.NO_COLOUR)

    def __init__(self, **kwargs):
        """
        Constructor
        """
        self.within             = None                      # An encapsulating object

        # Description
        self.str_desc           = ""

        # The main server function
        self.threaded_server    = None

        # The listener thread array -- each element of this array is threaded listener
        # object
        self.l_listener         = []
        self.listenerSleep      = 0.1

        # The fileIO threaded object
        self.fileIO             = None

        # DB
        self.b_clearDB          = False
        self.str_DBpath         = '/tmp/pman'
        self._ptree             = C_stree()
        self.str_fileio         = 'json'

        # Comms
        self.str_protocol       = "tcp"
        self.str_IP             = "127.0.0.1"
        self.str_port           = "5010"
        self.router_raw         = 0
        self.listeners          = 1
        self.b_http             = False

        # Job info
        self.auid               = ''
        self.jid                = ''

        # Debug parameters
        self.str_debugFile      = '/dev/null'
        self.b_debugToFile      = True

        for key,val in kwargs.items():
            if key == 'protocol':       self.str_protocol   = val
            if key == 'IP':             self.str_IP         = val
            if key == 'port':           self.str_port       = val
            if key == 'raw':            self.router_raw     = int(val)
            if key == 'listeners':      self.listeners      = int(val)
            if key == 'listenerSleep':  self.listenerSleep  = float(val)
            if key == 'http':           self.b_http         = int(val)
            if key == 'within':         self.within         = val
            if key == 'debugFile':      self.str_debugFile  = val
            if key == 'debugToFile':    self.b_debugToFile  = val
            if key == 'DBpath':         self.str_DBpath     = val
            if key == 'clearDB':        self.b_clearDB      = val
            if key == 'desc':           self.str_desc       = val

        # pudb.set_trace()

        # Screen formatting
        self.LC                 = 30
        self.RC                 = 50
        self.dp                 = debug(    verbosity   = 0,
                                            level       = -1,
                                            debugFile   = self.str_debugFile,
                                            debugToFile = self.b_debugToFile)

        if self.b_clearDB and os.path.isdir(self.str_DBpath):
            shutil.rmtree(self.str_DBpath)

        print(self.str_desc)

        # self.dp.qprint(Colors.YELLOW)
        # self.dp.qprint("""
        # \t+-----------------------------------------------+
        # \t| Welcome to the pman process management system |
        # \t+-----------------------------------------------+
        # """)
        # self.dp.qprint(Colors.CYAN + """
        # 'pman' is a client/server system that allows users to monitor
        # and control processes on (typically) Linux systems. Actual
        # processes are spawned using the 'crunner' module and as such
        # are ssh and HPC aware.
        #
        # The 'pman' server can be queried for running processes, lost/dead
        # processes, exit status, etc.
        #
        # Communication from the 'pman' server is via JSON constructs. See the
        # github page for more information.
        #
        # Typical calling syntax is:
        #
        #         ./pman.py   --raw 1                 \\
        #                     --http                  \\
        #                     --ip <someIP>           \\
        #                     --port 5010             \\
        #                     --listeners <listeners>
        #
        # """)

        # pudb.set_trace()
        self.col2_print('Server is listening on',
                        '%s://%s:%s' % (self.str_protocol, self.str_IP, self.str_port))
        self.col2_print('Router raw mode',                  str(self.router_raw))
        self.col2_print('HTTP response back mode',          str(self.b_http))
        self.col2_print('listener sleep',                   str(self.listenerSleep))

        # Read the DB from HDD
        self._ptree             = C_stree()
        # self.DB_read()
        self.DB_fileIO(cmd = 'load')

        # Setup zmq context
        self.zmq_context        = zmq.Context()

    def DB_read(self, **kwargs):
        """
        Read the DB from filesystem. If DB does not exist on filesystem,
        create an empty DB and save to filesystem.
        """
        if os.path.isdir(self.str_DBpath):
            self.dp.qprint("Reading pman DB from disk...\n")
            self._ptree = C_stree.tree_load(
                pathDiskRoot    = self.str_DBpath,
                loadJSON        = True,
                loadPickle      = False)
            self.dp.qprint("pman DB read from disk...\n")
            self.col2_print('Reading pman DB from disk:', 'OK')
        else:
            P = self._ptree
            # P.cd('/')
            # P.mkdir('proc')
            P.tree_save(
                startPath       = '/',
                pathDiskRoot    = self.str_DBpath,
                failOnDirExist  = False,
                saveJSON        = True,
                savePickle      = False
            )
            self.col2_print('Reading pman DB from disk:',
                            'No DB found... creating empty default DB')
        self.dp.qprint(Colors.NO_COLOUR, end='')

    def DB_fileIO(self, **kwargs):
        """
        Process DB file IO requests. Typically these control the
        DB -- save or load.
        """
        str_cmd     = 'save'
        str_DBpath  = self.str_DBpath
        str_fileio  = 'json'
        tree_DB     = self._ptree

        for k,v in kwargs.items():
            if k == 'cmd':      str_cmd             = v
            if k == 'fileio':   self.str_fileio     = v
            if k == 'dbpath':   str_DBpath          = v
            if k == 'db':       tree_DB             = v

        self.dp.qprint('cmd      = %s' % str_cmd)
        self.dp.qprint('fileio   = %s' % self.str_fileio)
        self.dp.qprint('dbpath   = %s' % str_DBpath)

        if str_cmd == 'save':
            if os.path.isdir(str_DBpath):
                shutil.rmtree(str_DBpath)
            #print(tree_DB)
            if self.str_fileio   == 'json':
                tree_DB.tree_save(
                    startPath       = '/',
                    pathDiskRoot    = str_DBpath,
                    failOnDirExist  = False,
                    saveJSON        = True,
                    savePickle      = False)
            if self.str_fileio   == 'pickle':
                tree_DB.tree_save(
                    startPath       = '/',
                    pathDiskRoot    = str_DBpath,
                    failOnDirExist  = False,
                    saveJSON        = False,
                    savePickle      = True)

        if str_cmd == 'load':
            if os.path.isdir(str_DBpath):
                self.dp.qprint("Reading pman DB from disk...\n")
                if self.str_fileio   == 'json':
                    tree_DB = C_stree.tree_load(
                        startPath       = '/',
                        pathDiskRoot    = str_DBpath,
                        failOnDirExist  = False,
                        loadJSON        = True,
                        loadPickle      = False)
                if self.str_fileio   == 'pickle':
                    tree_DB = C_stree.tree_load(
                        startPath       = '/',
                        pathDiskRoot    = str_DBpath,
                        failOnDirExist  = False,
                        loadJSON        = False,
                        loadPickle      = True)
                self.dp.qprint("pman DB read from disk...\n")
                self.col2_print('Reading pman DB from disk:', 'OK')
                self._ptree         = tree_DB
            else:
                tree_DB.tree_save(
                    startPath       = '/',
                    pathDiskRoot    = str_DBpath,
                    failOnDirExist  = False,
                    saveJSON        = True,
                    savePickle      = False
                )
                self.col2_print('Reading pman DB from disk:',
                                'No DB found... creating empty default DB')
            self.dp.qprint(Colors.NO_COLOUR, end='')

    def thread_serve(self):
        """
        Serve the 'start' method in a thread.
        :return:
        """
        self.threaded_server  = StoppableThread(target=self.start)
        self.threaded_server.start()

        while not self.threaded_server.stopped():
            time.sleep(1)

        # Stop the listeners...
        self.dp.qprint("setting b_stopThread on all listeners...")
        for i in range(0, self.listeners):
            self.dp.qprint("b_stopThread on listener %d and executing join()..." % i)
            self.l_listener[i].b_stopThread = True
            self.l_listener[i].join()

        # Stop the fileIO
        self.fileIO.b_stopThread    = True
        self.dp.qprint("b_stopThread on fileIO executing join()...")
        self.fileIO.join()

        self.dp.qprint("Shutting down the zmq infrastructure...")
        try:
            self.dp.qprint('calling self.socket_back.close()')
            self.socket_back.close()
        except:
            self.dp.qprint('Caught exception in closing back socket')

        try:
            self.dp.qprint('calling self.socket_front.close()')
            self.socket_front.close()
        except zmq.error.ZMQError:
            self.dp.qprint('Caught exception in closing front socket...')

        self.dp.qprint('calling zmq_context.term()')
        # self.zmq_context.term()

        self.dp.qprint("calling join() on all this thread...")
        self.threaded_server.join()
        self.dp.qprint("shutdown successful...")

    def start(self):
        """
            Main execution.

            * Instantiate several 'listener' worker threads
                **  'listener' threads are used to process input from external
                    processes. In turn, 'listener' threads can thread out
                    'crunner' threads that actually "run" the job.
            * Instantiate a job poller thread
                **  'poller' examines the internal DB entries and regularly
                    queries the system process table, tracking if jobs
                    are still running.
        """

        self.col2_print('Starting Listener threads', self.listeners)

        # Front facing socket to accept client connections.
        self.socket_front = self.zmq_context.socket(zmq.ROUTER)
        self.socket_front.router_raw = self.router_raw
        self.socket_front.setsockopt(zmq.LINGER, 1)
        self.socket_front.bind('%s://%s:%s' % (self.str_protocol,
                                          self.str_IP,
                                          self.str_port)
                          )

        # Backend socket to distribute work.
        self.socket_back = self.zmq_context.socket(zmq.DEALER)
        self.socket_back.setsockopt(zmq.LINGER, 1)
        self.socket_back.bind('inproc://backend')

        # Start the 'fileIO' thread
        self.fileIO      = FileIO(      timeout     = 60,
                                        within      = self,
                                        debugFile   = self.str_debugFile,
                                        debugToFile = self.b_debugToFile)
        self.fileIO.start()

        # Start the 'listener' workers... keep track of each
        # listener instance so that we can selectively stop
        # them later.
        for i in range(0, self.listeners):
            self.l_listener.append(Listener(
                                    id              = i,
                                    context         = self.zmq_context,
                                    DB              = self._ptree,
                                    DBpath          = self.str_DBpath,
                                    http            = self.b_http,
                                    within          = self,
                                    listenerSleep   = self.listenerSleep,
                                    debugToFile     = self.b_debugToFile,
                                    debugFile       = self.str_debugFile))
            self.l_listener[i].start()

        # Use built in queue device to distribute requests among workers.
        # What queue device does internally is,
        #   1. Read a client's socket ID and request.
        #   2. Send socket ID and request to a worker.
        #   3. Read a client's socket ID and result from a worker.
        #   4. Route result back to the client using socket ID.
        self.dp.qprint("*******before  zmq.device!!!")
        try:
            zmq.device(zmq.QUEUE, self.socket_front, self.socket_back)
        except:
            self.dp.qprint('Hmmm... some error was caught on shutting down the zmq.device...')
        self.dp.qprint("*******after zmq.device!!!")

    def __iter__(self):
        yield('Feed', dict(self._stree.snode_root))

    # @abc.abstractmethod
    # def create(self, **kwargs):
    #     """Create a new tree
    #
    #     """

    def __str__(self):
        """Print
        """
        return str(self.stree.snode_root)

    @property
    def stree(self):
        """STree Getter"""
        return self._stree

    @stree.setter
    def stree(self, value):
        """STree Getter"""
        self._stree = value

class FileIO(threading.Thread):
    """
    A class that periodically saves the database from memory out to disk.
    """

    def __init__(self, **kwargs):
        self.__name             = "FileIO"
        self.b_http             = False

        self.str_DBpath         = "/tmp/pman"

        self.timeout            = 60
        self.within             = None

        self.b_stopThread       = False

        # Debug parameters
        self.str_debugFile      = '/dev/null'
        self.b_debugToFile      = True

        for key,val in kwargs.items():
            if key == 'DB':             self._ptree         = val
            if key == 'DBpath':         self.str_DBpath     = val
            if key == 'timeout':        self.timeout        = val
            if key == 'within':         self.within         = val
            if key == 'debugFile':      self.str_debugFile  = val
            if key == 'debugToFile':    self.b_debugToFile  = val

        self.dp                 = debug(verbosity   = 0,
                                        level       = -1,
                                        debugFile   = self.str_debugFile,
                                        debugToFile = self.b_debugToFile)

        threading.Thread.__init__(self)

    def run(self):
        """ Main execution. """
        # pudb.set_trace()
        # Socket to communicate with front facing server.
        while not self.b_stopThread:
            # self.dp.qprint('Saving DB as type "%s" to "%s"...' % (
            #     self.within.str_fileio,
            #     self.within.str_DBpath
            # ))
            self.within.DB_fileIO(cmd = 'save')
            # self.dp.qprint('DB saved...')
            for second in range(0, self.timeout):
                if not self.b_stopThread:
                    time.sleep(1)
                else:
                    break

        self.dp.qprint('returning from FileIO run method...')
        # raise ValueError('FileIO thread terminated.')

class Listener(threading.Thread):
    """ Listeners accept communication requests from front facing server.
        Parse input text streams and act accordingly. """

    def __init__(self, **kwargs):
        self.__name             = "Listener"
        self.b_http             = False

        self.poller             = None
        self.str_DBpath         = "/tmp/pman"
        self.str_jobRootDir     = ''

        self.listenerSleep      = 0.1

        self.jid                = ''
        self.auid               = ''

        self.within             = None
        self.b_stopThread       = False

        # Debug parameters
        self.str_debugFile      = '/dev/null'
        self.b_debugToFile      = True

        for key,val in kwargs.items():
            if key == 'context':        self.zmq_context    = val
            if key == 'listenerSleep':  self.listenerSleep  = float(val)
            if key == 'id':             self.worker_id      = val
            if key == 'DB':             self._ptree         = val
            if key == 'DBpath':         self.str_DBpath     = val
            if key == 'http':           self.b_http         = val
            if key == 'within':         self.within         = val
            if key == 'debugFile':      self.str_debugFile  = val
            if key == 'debugToFile':    self.b_debugToFile  = val

        self.dp                 = debug(verbosity   = 0,
                                        level       = -1,
                                        debugFile   = self.str_debugFile,
                                        debugToFile = self.b_debugToFile)

        threading.Thread.__init__(self)
        # logging.debug('leaving __init__')

    def run(self):
        """ Main execution. """
        # Socket to communicate with front facing server.
        self.dp.qprint('starting...')
        socket = self.zmq_context.socket(zmq.DEALER)
        socket.connect('inproc://backend')

        b_requestWaiting        = False
        resultFromProcessing    = False
        request                 = ""
        client_id               = -1
        self.dp.qprint(Colors.BROWN + "Listener ID - %s: run() - Ready to serve..." % self.worker_id)
        while not self.b_stopThread:

            # wait (non blocking) for input on socket
            try:
                client_id, request  = socket.recv_multipart(flags = zmq.NOBLOCK)
                self.dp.qprint('Received %s from client_id: %s' % (request, client_id))
                b_requestWaiting    = True
            except zmq.Again as e:
                if self.listenerSleep:
                    time.sleep(0.1)
                else:
                    pass

            if b_requestWaiting:
                self.dp.qprint(Colors.BROWN + 'Listener ID - %s: run() - Received comms from client.' % (self.worker_id))
                self.dp.qprint(Colors.BROWN + 'Client sends: %s' % (request))

                resultFromProcessing    = self.process(request)
                if resultFromProcessing:
                    self.dp.qprint(Colors.BROWN + 'Listener ID - %s: run() - Sending response to client.' %
                                   (self.worker_id))
                    pudb.set_trace()
                    self.dp.qprint('JSON formatted response:')
                    str_payload = json.dumps(resultFromProcessing)
                    self.dp.qprint(Colors.LIGHT_CYAN + str_payload)
                    self.dp.qprint(Colors.BROWN + 'len = %d chars' % len(str_payload))
                    socket.send(client_id, zmq.SNDMORE)
                    if self.b_http:
                        str_contentType = "application/json"
                        res  = Response(str_payload)
                        res.content_type = str_contentType

                        str_HTTPpre = "HTTP/1.x "
                        str_res     = "%s%s" % (str_HTTPpre, str(res))
                        str_res     = str_res.replace("UTF-8", "UTF-8\nAccess-Control-Allow-Origin: *")

                        socket.send(str_res.encode())
                    else:
                        socket.send(str_payload)
            b_requestWaiting    = False
        self.dp.qprint('Listnener ID - %s: Returning from run()...' % self.worker_id)
        # raise('Listener ID - %s: Thread terminated' % self.worker_id)
        return True

    def t_search_process(self, *args, **kwargs):
        """

        Search

        :param args:
        :param kwargs:
        :return:
        """

        self.dp.qprint("In search process...")

        d_request   = {}
        d_ret       = {}
        hits        = 0

        for k, v in kwargs.items():
            if k == 'request':      d_request   = v

        d_meta          = d_request['meta']

        b_pathSpec      = False
        str_path        = ""
        if 'path' in d_meta:
            b_pathSpec  = True
            str_path    = d_meta['path']

        b_jobSpec       = False
        str_jobSpec    = ""
        if 'job' in d_meta:
            b_jobSpec   = True
            str_jobSpec = d_meta['job']

        b_fieldSpec    = False
        str_fieldSpec  = ""
        if 'field' in d_meta:
            b_fieldSpec = True
            str_fieldSpec = d_meta['field']

        b_whenSpec      = False
        str_whenSpec    = "end"
        if 'when' in d_meta:
            b_whenSpec = True
            str_whenSpec = d_meta['when']

        self.dp.qprint(d_meta)
        self.dp.qprint(b_pathSpec)
        str_fileName    = d_meta['key']
        str_target      = d_meta['value']
        p               = self._ptree
        str_origDir     = p.cwd()
        str_pathOrig    = str_path
        for r in self._ptree.lstr_lsnode('/'):
            if p.cd('/' + r)['status']:
                str_val = p.cat(str_fileName)
                if str_val == str_target:
                    if not b_pathSpec:
                        str_path            = '/api/v1/' + r + '/' + str_fileName
                    else:
                        str_path            = '/api/v1/' + r + str_pathOrig
                        if str_path[-1] == '/': str_path = str_path[:-1]
                    if b_jobSpec:
                        str_path            = '/api/v1/' + r +              '/' + \
                                                str_whenSpec +              '/' + \
                                                str_jobSpec +               '/' + \
                                                '%sInfo' % str_whenSpec +   '/' + \
                                                str_jobSpec +               '/' + \
                                                str_fieldSpec
                    d_ret[str(hits)]    = {}
                    d_ret[str(hits)]    = self.DB_get(path = str_path)
                    hits               += 1
        p.cd(str_origDir)

        return {"d_ret":    d_ret,
                "status":   bool(hits)}

    def t_info_process(self, *args, **kwargs):
        """

        Check if the job corresponding to the search pattern is "done".

        :param args:
        :param kwargs:
        :return:
        """

        self.dp.qprint("In info process...")

        d_request   = {}
        d_ret       = {}
        b_status    = False
        hits        = 0
        for k, v in kwargs.items():
            if k == 'request':      d_request   = v

        d_search    = self.t_search_process(request = d_request)['d_ret']

        p = self._ptree
        for j in d_search.keys():
            d_j = d_search[j]
            for job in d_j.keys():
                str_pathStart       = '/api/v1/' + job + '/startInfo'
                str_pathEnd         = '/api/v1/' + job + '/endInfo'
                d_ret[str(hits)+'.0']    = {}
                d_ret[str(hits)+'.0']    = self.DB_get(path = str_pathStart)
                d_ret[str(hits)+'.1']    = {}
                d_ret[str(hits)+'.1']    = self.DB_get(path = str_pathEnd)
                hits               += 1
        if not hits:
            d_ret                   = {
                "-1":   {
                    "noJobFound":   {
                        "endInfo":  {"allJobsDone": None}
                    }
                }
            }
        else:
            b_status            = True
        return {"d_ret":    d_ret,
                "status":   b_status}

    def t_quit_process(self, *args, **kwargs):
        """
        Process the 'quit' POST directive. This might appear counter-inuitive
        at first glance since the 'get' is the result of a REST POST, but is
        logically consistent within the semantics of this system.
        """
        d_request   = {}
        d_ret       = {}
        b_status    = False
        hits        = 0
        for k, v in kwargs.items():
            if k == 'request':      d_request   = v
        d_meta      = d_request['meta']
        if 'saveDB' in d_meta.keys():
            self.dp.qprint("Saving DB...")
            self.within.DB_fileIO(cmd = 'save')

        self.dp.qprint('calling threaded_server.stop()')
        self.within.threaded_server.stopit()
        self.dp.qprint('called threaded_server.stop()')

        return {'d_ret':    d_ret,
                'status':   True}

    def t_get_process(self, *args, **kwargs):
        """
        Process the 'get' POST directive. This might appear counter-inuitive
        at first glance since the 'get' is the result of a REST POST, but is
        logically consistent within the semantics of this system.
        """
        d_request   = {}
        d_ret       = {}
        b_status    = False
        hits        = 0
        for k, v in kwargs.items():
            if k == 'request':      d_request   = v
        d_meta      = d_request['meta']
        str_path    = '/api/v1' + d_meta['path']
        d_ret       = self.DB_get(path  = str_path)
        return {'d_ret':    d_ret,
                'status':   True}

    def t_fileiosetup_process(self, *args, **kwargs):
        """
        Setup a thread with a socket listener. Return listener address to client
        """
        self.dp.qprint("In fileiosetup process...")

        d_ret               = {}
        for k, v in kwargs.items():
            if k == 'request':      d_request   = v

        d_meta  = d_request['meta']

        d_ret['fileioIP']   = "%s" % self.within.str_IP
        d_ret['fileioport'] = "%s" % (int(self.within.str_port) + self.worker_id)
        d_ret['serveforever']=d_meta['serveforever']

        d_args              = {
                                'ip':    d_ret['fileioIP'],
                                'port':  d_ret['fileioport']
                               }

        server              = ThreadedHTTPServer((d_args['ip'], int(d_args['port'])), StoreHandler)
        server.setup(args   = d_args)
        self.dp.qprint("serveforever = %d" % d_meta['serveforever'])
        b_serveforever      = False
        if 'serveforever' in d_meta.keys():
            b_serveforever  = d_meta['serveforever']

        if b_serveforever:
            self.dp.qprint("about to serve_forever()...")
            server.serve_forever()
        else:
            self.dp.qprint("about to handle_request()...")
            server.handle_request()

        return {"d_ret":    d_ret,
                "status":   True}

    def job_state(self, *args, **kwargs):
        """

        Return a structure that can be further processed to determine the job's state.

        :param args:
        :param kwargs:
        :return:
        """

        self.dp.qprint("In done process...")

        d_request   = {}
        d_ret       = {}
        b_status    = False
        b_container = False
        hits        = 0
        for k, v in kwargs.items():
            if k == 'request':      d_request   = v

        d_search    = self.t_search_process(request = d_request)['d_ret']

        p   = self._ptree
        Ts  = C_stree()
        Te  = C_stree()
        for j in d_search.keys():
            d_j = d_search[j]
            for job in d_j.keys():
                str_pathStart       = '/api/v1/' + job + '/start'
                str_pathEnd         = '/api/v1/' + job + '/end'

                d_start             = self.DB_get(path = str_pathStart)
                d_end               = self.DB_get(path = str_pathEnd)
                Ts.initFromDict(d_start)
                Te.initFromDict(d_end)

                self.dp.qprint("Ts.cwd = %s " % Ts.cwd())
                self.dp.qprint(Ts)
                self.dp.qprint("Te.cwd = %s " % Te.cwd())
                self.dp.qprint(Te)

                l_subJobsStart      = []
                if Ts.cd('/%s/start' % job)['status']:
                    l_subJobsStart  = Ts.lstr_lsnode()
                    l_subJobsStart  = list(map(int, l_subJobsStart))
                    l_subJobsStart.sort()
                    self.dp.qprint("l_subJobsStart  (pre) = %s" % l_subJobsStart)
                    if len(l_subJobsStart) > 1: l_subJobsStart  = l_subJobsStart[:-1]

                l_subJobsEnd        = []
                if Te.cd('/%s/end' % job)['status']:
                    l_subJobsEnd    = Te.lstr_lsnode()
                    l_subJobsEnd    = list(map(int, l_subJobsEnd))
                    l_subJobsEnd.sort()
                    self.dp.qprint("l_subJobsEnd    (pre) = %s " % l_subJobsEnd)
                    if len(l_subJobsEnd) > 1: l_subJobsEnd    = l_subJobsEnd[:-1]

                self.dp.qprint("l_subJobsStart (post) = %s" % l_subJobsStart)
                self.dp.qprint("l_subJobsEnd   (post) = %s" % l_subJobsEnd)

                for j in l_subJobsStart:
                    l_subJobsStart[j]   = Ts.cat('/%s/start/%d/startInfo/%d/startTrigger' % \
                                                 (job, j, j))

                # jobsEnd behaviour can be slightly different to the jobStart, particularly if
                # the job being executed is killed -- sometimes recording the "death" event of
                # the job does not happen and the job indexing ends up missing several epochs:
                #
                #           l_subJobsStart  (pre) = [0, 1, 2, 3, 4]
                #           l_subJobsEnd    (pre) = [0, 1, 3, 4]
                #
                # to assure correct returncode lookup, we always parse the latest job epoch.

                latestJob       = 0
                if len(l_subJobsEnd):
                    latestJob   = l_subJobsEnd[-1]
                    for j in list(range(0, latestJob+1)):
                        l_subJobsEnd[j]         = Te.cat('/%s/end/%s/endInfo/%d/returncode' % (job, latestJob, j))
                T_container     = False
                if p.exists('container', path = '/%s' % job):
                    T_container = C_stree()
                    p.copy(startPath = '/%s/container' % job, destination = T_container)
                d_ret[str(hits)+'.start']       = {"jobRoot": job, "startTrigger":  l_subJobsStart}
                d_ret[str(hits)+'.end']         = {"jobRoot": job, "returncode":    l_subJobsEnd}
                d_ret[str(hits)+'.container']   = {"jobRoot": job, "tree":          dict(T_container.snode_root)}
                hits               += 1
        if not hits:
            d_ret['-1.start']       = {"jobRoot":   None, "startTrigger":   None}
            d_ret['-1.end']         = {"jobRoot":   None, "returncode":     None}
            d_ret['-1.container']   = {"jobRoot":   None, "tree":           None}
        else:
            b_status            = True
        return {"hits":         hits,
                "d_ret":        d_ret,
                "status":       b_status}

    def t_done_process(self, *args, **kwargs):
        """

        Check if the job corresponding to the search pattern is "done".

        :param args:
        :param kwargs:
        :return:
        """

        self.dp.qprint("In done process...")

        return self.job_state(*args, **kwargs)

    def t_status_process(self, *args, **kwargs):
        """

        Return status on a given job.

        Attempts to check if a containerized job, and branch accordingly.

        :param args:
        :param kwargs:
        :return:
        """

        self.dp.qprint("In status process...")

        d_state     = self.job_state(*args, **kwargs)
        d_ret       = d_state['d_ret']
        b_status    = d_state['status']

        l_keys      = d_ret.items()
        l_status    = []
        for i in range(0, int(len(l_keys)/2)):
            b_startEvent    = d_ret['%s.start'  % str(i)]['startTrigger'][0]
            try:
                endcode     = d_ret['%s.end'    % str(i)]['returncode'][0]
            except:
                endcode     = None

            pudb.set_trace()
            if d_state['d_ret']['%s.container' % str(i)]['tree']:
                kwargs['d_state']   = d_state
                kwargs['hitIndex']  = str(i)
                l_status.append(self.t_status_process_container(*args, **kwargs))
            else:
                if endcode is None and b_startEvent:
                    l_status.append('started')
                if not endcode and b_startEvent and type(endcode) is int:
                    l_status.append('finishedSuccessfully')
                if endcode and b_startEvent:
                    l_status.append('finishedWithError')

            self.dp.qprint('b_startEvent = %d' % b_startEvent)
            self.dp.qprint(endcode)
            self.dp.qprint('l_status = %s' % l_status)

        d_ret['l_status']   = l_status

        return {"d_ret":    d_ret,
                "status":   b_status}

    def t_status_process_container_stateObject(self, *args, **kwargs):
        """
        Process the actual JSON container return object on service
        state.

        PRECONDITIONS:
        o   This method should only ever be called by t_status_process_container().

        POSTCONDITIONS:
        o   A string denoting the current state is returned.
        o   If state is complete and service still running, save state object to
            tree and remove service.

        """
        d_serviceState      = None
        d_jobState          = None
        str_hitIndex        = "0"
        str_ret             = 'undefined'
        b_shutDownService   = False
        for k,v in kwargs.items():
            if k == 'jobState':         d_jobState      = v
            if k == 'serviceState':     d_serviceState  = v
            if k == 'hitIndex':         str_hitIndex    = v
        if d_serviceState:
            str_state   = d_serviceState['Status']['State']
            str_message = d_serviceState['Status']['Message']
            if str_state == 'running'   and str_message == 'started':
                str_ret = 'started'
            if str_state == 'failed'    and str_message == 'started':
                str_ret = 'finishedWithError'
                b_shutDownService   = True
            if str_state == 'complete'  and str_message == 'finished':
                str_ret     = 'finishedSuccessfully'
                b_shutDownService   = True
            if b_shutDownService:
                str_jobRoot = d_jobState['d_ret']['%s.container' % str_hitIndex]['jobRoot']
                if not self._ptree.exists('state', path = '/%s/container' % str_jobRoot):
                    # Store the service state
                    self._ptree.touch('/%s/container/state' % str_jobRoot, d_serviceState)
                    # And remove the service from the container scheduler
                    client  = docker.from_env()
                    self._ptree.cd('/%s/container' % str_jobRoot)
                    str_serviceName     = self._ptree.cat('manager/env/serviceName')
                    str_managerImage    = self._ptree.cat('manager/image')
                    str_managerApp      = self._ptree.cat('manager/app')
                    str_cmdManager      = '%s --remove %s' % \
                                          (str_managerApp, str_serviceName)
                    byte_str = client.containers.run('%s' % str_managerImage,
                                                     str_cmdManager,
                                                     volumes={'/var/run/docker.sock': {'bind': '/var/run/docker.sock', 'mode': 'rw'}},
                                                     remove=True)
        return str_ret

    def t_status_process_container(self, *args, **kwargs):
        """
        Determine the status of a job scheduled using the container manager.

        PRECONDITIONS:
        o   Only call this method if a container structure exists
            in the relevant job tree!

        POSTCONDITIONS:
        o   If the job is completed, then shutdown the container cluster
            service.

        """
        d_state         = None
        str_jobRoot     = ''
        str_hitIndex    = "0"

        for k,v in kwargs.items():
            if k == 'd_state':  d_state         = v
            if k == 'hitIndex': str_hitIndex    = v

        self.dp.qprint('checking on status using container...')
        str_jobRoot         = d_state['d_ret']['%s.container' % str_hitIndex]['jobRoot']
        self._ptree.cd('/%s/container' % str_jobRoot)
        str_serviceName     = self._ptree.cat('manager/env/serviceName')
        str_managerImage    = self._ptree.cat('manager/image')
        str_managerApp      = self._ptree.cat('manager/app')

        # Check if the state of the container service has been recorded to the data tree
        if self._ptree.exists('state', path = '/%s/container' % str_jobRoot):
            # The job has actually completed and its state recorded in the data tree
            d_json          = self._ptree.cat('/%s/container/state')
        else:
            # Ask the container service for the state of the service...
            client = docker.from_env()

            # Get the state of the service...
            str_cmdManager  = '%s --state %s' % \
                              (str_managerApp, str_serviceName)
            byte_str        = client.containers.run('%s' % str_managerImage,
                                                    str_cmdManager,
                                                    volumes = {'/var/run/docker.sock': {'bind': '/var/run/docker.sock', 'mode': 'rw'}},
                                                    remove  = True)
            d_json          = json.loads(byte_str.decode())
        return self.t_status_process_container_stateObject( hitIndex        = str_hitIndex,
                                                            jobState        = d_state,
                                                            serviceState    = d_json)

    def t_hello_process(self, *args, **kwargs):
        """

        The 'hello' action is merely to 'speak' with the server. The server
        can return current date/time, echo back a string, query the startup
        command line args, etc.

        This method is a simple means of checking if the server is "up" and
        running.

        :param args:
        :param kwargs:
        :return:
        """

        self.dp.qprint("In hello process...")
        b_status            = False
        d_ret               = {}
        for k, v in kwargs.items():
            if k == 'request':      d_request   = v

        d_meta  = d_request['meta']
        if 'askAbout' in d_meta.keys():
            str_askAbout    = d_meta['askAbout']
            if str_askAbout == 'timestamp':
                str_timeStamp   = datetime.datetime.today().strftime('%Y%m%d%H%M%S.%f')
                d_ret['timestamp']              = {}
                d_ret['timestamp']['now']       = str_timeStamp
                b_status                        = True
            if str_askAbout == 'sysinfo':
                d_ret['sysinfo']                = {}
                d_ret['sysinfo']['system']      = platform.system()
                d_ret['sysinfo']['machine']     = platform.machine()
                d_ret['sysinfo']['platform']    = platform.platform()
                d_ret['sysinfo']['uname']       = platform.uname()
                d_ret['sysinfo']['version']     = platform.version()
                d_ret['sysinfo']['memory']      = psutil.virtual_memory()
                d_ret['sysinfo']['cpucount']    = multiprocessing.cpu_count()
                d_ret['sysinfo']['loadavg']     = os.getloadavg()
                d_ret['sysinfo']['cpu_percent'] = psutil.cpu_percent()
                d_ret['sysinfo']['hostname']    = socket.gethostname()
                b_status                        = True
            if str_askAbout == 'echoBack':
                d_ret['echoBack']               = {}
                d_ret['echoBack']['msg']        = d_meta['echoBack']
                b_status                        = True

        return { 'd_ret':   d_ret,
                 'status':  b_status}

    def t_run_process(self, *args, **kwargs):
        """
        Main job handler -- this is in turn a thread spawned from the
        parent listener thread.
        By being threaded, the client http caller gets an immediate
        response without needing to wait on the jobs actually running
        to completion.
        """

        str_cmd             = ""
        d_request           = {}
        d_meta              = {}
        d_Tcontainer        = {}

        for k,v in kwargs.items():
            if k == 'request':      d_request       = v
            if k == 'treeList':     d_Tcontainer    = v

        d_meta          = d_request['meta']

        if d_meta:
            self.jid    = d_meta['jid']
            self.auid   = d_meta['auid']
            str_cmd     = d_meta['cmd']

        if isinstance(self.jid, int):
            self.jid    = str(self.jid)

        self.dp.qprint("spawing and starting poller thread")

        # Start the 'poller' worker
        self.poller  = Poller(cmd           = str_cmd,
                              debugToFile   = self.b_debugToFile,
                              debugFile     = self.str_debugFile)
        self.poller.start()

        str_timeStamp       = datetime.datetime.today().strftime('%Y%m%d%H%M%S.%f')
        str_uuid            = uuid.uuid4()
        str_dir             = '%s_%s' % (str_timeStamp, str_uuid)
        self.str_jobRootDir = str_dir

        b_jobsAllDone       = False
        p                   = self._ptree

        p.cd('/')
        p.mkcd(str_dir)

        if d_Tcontainer:
            # Save the trees in this list to the DB...
            for name,tree in d_Tcontainer.items():
                p.mkcd(name)
                tree.copy(startPath = '/', destination = p, pathDiskRoot = '/%s/%s' % (str_dir, name))
                p.cd('/%s' % str_dir)

        p.touch('d_meta',       json.dumps(d_meta))
        p.touch('cmd',          str_cmd)
        if len(self.auid):
            p.touch('auid',     self.auid)
        if len(self.jid):
            p.touch('jid',      self.jid)

        p.mkdir('start')
        p.mkdir('end')

        jobCount        = 0
        p.touch('jobCount',     jobCount)

        while not b_jobsAllDone:
            try:
                b_jobsAllDone   = self.poller.queueAllDone.get_nowait()
            except queue.Empty:
                self.dp.qprint('Waiting on start job info')
                d_startInfo     = self.poller.queueStart.get()
                str_startDir    = '/%s/start/%d' % (self.str_jobRootDir, jobCount)
                p.mkdir(str_startDir)
                p.cd(str_startDir)
                p.touch('startInfo', d_startInfo.copy())
                p.touch('/%s/startInfo' % str_dir, d_startInfo.copy())

                self.dp.qprint('Waiting on end job info')
                d_endInfo       = self.poller.queueEnd.get()
                str_endDir      = '/%s/end/%d' % (self.str_jobRootDir, jobCount)
                p.mkdir(str_endDir)
                p.cd(str_endDir)
                p.touch('endInfo', d_endInfo.copy())
                p.touch('/%s/endInfo' % str_dir,    d_endInfo.copy())

                p.touch('/%s/jobCount' % str_dir,   jobCount)
                jobCount        += 1
        self.dp.qprint('All jobs processed.')

        # Save DB state...
        self.within.DB_fileIO(cmd = 'save')

    def t_run_process_container(self, *args, **kwargs):
        """
        A threaded run method specialized to handling containerized managers and targets.

        NOTE: If 'serviceName' is not specified/present, then this defaults to the 'jid'
        value and is in fact the default behaviour.

        Typical JSON d_request:
        
        {   "action": "run",
            "meta":  {
                "cmd":      "$execshell $selfpath/$selfexec --prefix test --sleepLength 0 /share/incoming /share/outgoing",
                "auid":     "rudolphpienaar",
                "jid":      "simpledsapp-1",
                "threaded": true,
                "container":   {
                        "target": {
                            "image":     "fnndsc/pl-simpledsapp",
                            "cmdParse": true
                        },
                        "manager": {
                            "image":    "fnndsc/swarm"
                            "app":      "swarm.py",
                            "env":  {
                                "shareDir": "/home/tmp/share",
                                "serviceName":  "testService"
                            }
                        }
                }
            }
        }
        """

        str_cmd             = ""
        str_shareDir        = ""
        str_serviceName     = ""
        d_request           = {}
        d_meta              = {}
        d_container         = {}
        d_image             = {}
        d_manager           = {}
        d_env               = {}

        self.dp.qprint('Processing swarm-type job...')

        for k,v in kwargs.items():
            if k == 'request': d_request    = v

        d_meta          = d_request['meta']
        # pudb.set_trace()

        if d_meta:
            self.jid            = d_meta['jid']
            self.auid           = d_meta['auid']
            str_cmd             = d_meta['cmd']
            str_serviceName     = self.jid

            if 'container' in d_meta.keys():
                d_container         = d_meta['container']
                d_target            = d_container['target']
                str_targetImage     = d_target['image']

                d_manager           = d_container['manager']
                str_managerImage    = d_manager['image']
                str_managerApp      = d_manager['app']

                d_env               = d_manager['env']
                if 'shareDir' in d_env.keys():
                    str_shareDir    = d_env['shareDir']
                if 'serviceName' in d_env.keys():
                    str_serviceName = d_env['serviceName']

            # First, attach to the docker daemon...
            client = docker.from_env()

            #
            # If 'container/cmdParse', get a JSON representation of the image and
            # parse the cmd for substitutions -- this replaces any of
            # $exeshell, $selfpath, and/or $selfexec with the values provided
            # in the JSON representation.
            #
            if d_target['cmdParse']:
                byte_str    = client.containers.run(str_targetImage)
                d_jsonRep   = json.loads(byte_str.decode())
                for str_meta in ['execshell', 'selfexec', 'selfpath']:
                    str_cmd = str_cmd.replace("$"+str_meta, d_jsonRep[str_meta])

            str_cmdLine     = str_cmd
            str_cmdManager  = '%s -s %s -m %s -i %s -p none -c "%s"' % \
                              (str_managerApp, str_serviceName, str_shareDir, str_targetImage, str_cmdLine)
            byte_str        = client.containers.run('%s' % str_managerImage,
                                             str_cmdManager,
                                             volumes = {'/var/run/docker.sock': {'bind': '/var/run/docker.sock', 'mode': 'rw'}},
                                             remove  = True)

            # Call the "parent" method -- reset the cmdLine to an "echo"
            # and create an stree off the 'container' dictionary to store
            # in the pman DB entry.
            d_meta['cmd']   = 'echo "%s"' % str_cmd
            T_container     = C_stree()
            T_container.initFromDict(d_container)
            d_Tcontainer    = {'container': T_container}
            self.t_run_process(request  = d_request,
                               treeList = d_Tcontainer)
            self.dp.qprint('Returning from swarm-type job...')

    def json_filePart_get(self, **kwargs):
        """
        If the requested path is *within* a json "file" on the
        DB, then we need to find the file, and map the relevant
        path to components in that file.
        """

    def DB_get(self, **kwargs):
        """
        Returns part of the DB tree based on path spec in the URL
        """

        r           = C_stree()
        p           = self._ptree

        pcwd        = p.cwd()
        str_URLpath = "/api/v1/"
        for k,v in kwargs.items():
            if k == 'path':     str_URLpath = v

        str_path    = '/' + '/'.join(str_URLpath.split('/')[3:])

        self.dp.qprint("path = %s" % str_path)

        if str_path == '/':
            # If root node, only return list of jobs
            l_rootdir = p.lstr_lsnode(str_path)
            r.mknode(l_rootdir)
        else:
            # Here is a hidden behaviour. If the 'root' dir starts
            # with an underscore, then replace that component of
            # the path with the actual name in list order.
            # This is simply a short hand way to access indexed
            # offsets.

            l_path  = str_path.split('/')
            jobID   = l_path[1]
            # Does the jobID start with an underscore?
            if jobID[0] == '_':
                jobOffset   = jobID[1:]
                l_rootdir   = list(p.lstr_lsnode('/'))
                self.dp.qprint('jobOffset = %s' % jobOffset)
                self.dp.qprint(l_rootdir)
                try:
                    actualJob   = l_rootdir[int(jobOffset)]
                except:
                    return False
                l_path[1]   = actualJob
                str_path    = '/'.join(l_path)

            r.mkdir(str_path)
            r.cd(str_path)
            r.cd('../')
            # if not r.graft(p, str_path):
            # pudb.set_trace()
            if not p.copy(startPath = str_path, destination = r)['status']:
                # We are probably trying to access a file...
                # First, remove the erroneous path in the return DB
                r.rm(str_path)

                # Now, we need to find the "file", parse the json layer
                # and save...
                n                   = 0
                contents            = p.cat(str_path)
                str_pathFile        = str_path
                l_path              = str_path.split('/')
                totalPathLen        = len(l_path)
                l_pathFile          = []
                while not contents and -1*n < totalPathLen:
                    n               -= 1
                    str_pathFile    = '/'.join(str_path.split('/')[0:n])
                    contents        = p.cat(str_pathFile)
                    l_pathFile.append(l_path[n])

                if contents and n<0:
                    l_pathFile      = l_pathFile[::-1]
                    str_access      = ""
                    for l in l_pathFile:
                        str_access += "['%s']" % l
                    self.dp.qprint('str_access = %s' % str_access)
                    try:
                        contents        = eval('contents%s' % str_access)
                    except:
                        contents        = False

                r.touch(str_path, contents)

        p.cd(pcwd)

        self.dp.qprint(r)
        self.dp.qprint(dict(r.snode_root))
        return dict(r.snode_root)

        # return r

    def process(self, request, **kwargs):
        """ Process the message from remote client

        In some philosophical respects, this process() method in fact implements
        REST-like API of its own.

        """

        if len(request):

            REST_header     = ""
            REST_verb       = ""
            str_path        = ""
            json_payload    = ""

            self.dp.qprint("Listener ID - %s: process() - handling request" % (self.worker_id))

            now             = datetime.datetime.today()
            str_timeStamp   = now.strftime('%Y-%m-%d %H:%M:%S.%f')
            self.dp.qprint(Colors.YELLOW)
            self.dp.qprint("***********************************************")
            self.dp.qprint("***********************************************")
            self.dp.qprint("%s incoming data stream" % (str_timeStamp) )
            self.dp.qprint("***********************************************")
            self.dp.qprint("len = %d" % len(request))
            self.dp.qprint("***********************************************")
            self.dp.qprint(Colors.CYAN + "%s\n" % (request.decode()) + Colors.YELLOW)
            self.dp.qprint("***********************************************" + Colors.NO_COLOUR)
            l_raw           = request.decode().split('\n')
            FORMtype        = l_raw[0].split('/')[0]

            self.dp.qprint('Request = ...')
            self.dp.qprint(l_raw)
            REST_header             = l_raw[0]
            REST_verb               = REST_header.split()[0]
            str_path                = REST_header.split()[1]
            json_payload            = l_raw[-1]

            # remove trailing '/' if any on path
            if str_path[-1]         == '/': str_path = str_path[0:-1]

            d_ret                   = {'status': False,
                                       'RESTheader': REST_header,
                                       'RESTverb': REST_verb,
                                       'action': "",
                                       'path': str_path,
                                       'receivedByServer': l_raw}

            if REST_verb == 'GET':
                d_ret['GET']    = self.DB_get(path = str_path)
                d_ret['status'] = True

            self.dp.qprint('json_payload = %s' % json_payload)
            d_ret['client_json_payload']    = json_payload
            d_ret['client_json_len']        = len(json_payload)
            if len(json_payload):
                d_payload           = json.loads(json_payload)
                d_request           = d_payload['payload']
                payload_verb        = d_request['action']
                if 'meta' in d_request.keys():
                    d_meta          = d_request['meta']
                d_ret['payloadsize']= len(json_payload)

                if payload_verb == 'quit':
                    self.dp.qprint('Shutting down server...')
                    d_ret['status'] = True

                if payload_verb == 'run' and REST_verb == 'PUT':
                    d_ret['action']     = payload_verb
                    self.processPUT(    request     = d_request)
                    d_ret['status'] = True

                if REST_verb == 'POST':
                    self.processPOST(   request = d_request,
                                        ret     = d_ret)

            return d_ret
        else:
            return False
        
    def methodName_parse(self, **kwargs):
        """
        Construct the processing method name (string) by parsing the
        d_meta dictionary.
        """
        d_meta              = {}
        d_container         = {}
        str_method          = ""        # The main 'parent' method
        str_methodSuffix    = ""        # A possible 'subclass' specialization

        for k,v in kwargs.items():
            if k == 'request': d_request= v
        payload_verb        = d_request['action']

        if 'meta' in d_request.keys():
            d_meta          = d_request['meta']

        if 'container' in d_meta.keys():
            # If the 'container' json paragraph exists, then route processing to
            # a suffixed '_container' method.
            str_methodSuffix    = 'container'
            # d_container         = d_meta['container']
            # str_methodSuffix    = d_container['manager']['type']

        str_method  = 't_%s_process_%s' %(payload_verb, str_methodSuffix)
        return str_method

    def processPOST(self, **kwargs):
        """
         Dispatcher for POST
        """

        for k,v in kwargs.items():
            if k == 'request':  d_request   = v
            if k == 'ret':      d_ret       = v

        payload_verb        = d_request['action']
        if 'meta' in d_request.keys():
            d_meta          = d_request['meta']

        d_ret['action'] = payload_verb
        d_ret['meta']   = d_meta

        b_threaded      = False
        if 'threaded' in d_meta.keys():
            b_threaded  = d_meta['threaded']

        if b_threaded:
            self.dp.qprint("Will process request in new thread.")
            pf_method   = None
            # pudb.set_trace()
            str_method  = self.methodName_parse(request = d_request)
            # str_method  = 't_%s_process' % payload_verb
            try:
                pf_method  = getattr(self, str_method)
            except AttributeError:
                raise NotImplementedError("Class `{}` does not implement `{}`".format(my_cls.__class__.__name__, method_name))

            t_process           = threading.Thread(     target      = pf_method,
                                                        args        = (),
                                                        kwargs      = kwargs)
            t_process.start()
            time.sleep(0.1)
            if payload_verb == 'run':
                d_ret['jobRootDir'] = self.str_jobRootDir
            d_ret['status']     = True
        else:
            self.dp.qprint("Will process request in current thread.")
            d_done              = eval("self.t_%s_process(request = d_request)" % payload_verb)
            try:
                d_ret['d_ret']      = d_done["d_ret"]
                d_ret['status']     = d_done["status"]
            except:
                self.dp.qprint("An error occurred in reading ret structure. Should this method have been threaded?")

        return d_ret

    def processPUT(self, **kwargs):
        """
         Dispatcher for PUT
        """

        d_request       = {}
        str_action      = "run"
        str_cmd         = "save"
        str_DBpath      = self.str_DBpath
        str_fileio      = "json"
        tree_DB         = self._ptree

        for k,v in kwargs.items():
            if k == 'request':  d_request   = v

        str_action      = d_request['action']
        self.dp.qprint('action = %s' % str_action)
        d_meta              = d_request['meta']
        self.dp.qprint('action = %s' % str_action)

        # Optional search criteria
        if 'key'        in d_meta:
            d_search    = self.t_search_process(request = d_request)['d_ret']

            p           = self._ptree
            Tj          = C_stree()
            Tdb         = C_stree()
            for j in d_search.keys():
                d_j = d_search[j]
                for job in d_j.keys():
                    str_pathJob         = '/api/v1/' + job

                    d_job               = self.DB_get(path = str_pathJob)
                    Tj.initFromDict(d_job)
                    Tj.copy(startPath = '/', destination = Tdb)

                    # Tdb.graft(Tj, '/')

                    # self.DB_get(path = str_pathJob).copy(startPath = '/', destination = Tdb)


            # print(Tdb)
            tree_DB     = Tdb

        if 'context'    in d_meta:  str_context     = d_meta['context']
        if 'operation'  in d_meta:  str_cmd         = d_meta['operation']
        if 'dbpath'     in d_meta:  str_DBpath      = d_meta['dbpath']
        if 'fileio'     in d_meta:  str_type        = d_meta['fileio']

        if str_action.lower() == 'run' and str_context.lower() == 'db':
            self.within.DB_fileIO(  cmd         = str_cmd,
                                    fileio      = str_fileio,
                                    dbpath      = str_DBpath,
                                    db          = tree_DB)

class Poller(threading.Thread):
    """
    The Poller checks for running processes based on the internal
    DB and system process table. Jobs that are no longer running are
    removed from the internal DB.
    """

    def __init__(self, **kwargs):

        self.pollTime           = 10
        self.str_cmd            = ""
        self.crunner            = None
        self.queueStart         = queue.Queue()
        self.queueEnd           = queue.Queue()
        self.queueAllDone       = queue.Queue()

        # self.dp.qprint('starting...', level=-1)

        # Debug parameters
        self.str_debugFile      = '/dev/null'
        self.b_debugToFile      = True

        for key,val in kwargs.items():
            if key == 'pollTime':       self.pollTime       = val
            if key == 'cmd':            self.str_cmd        = val
            if key == 'debugFile':      self.str_debugFile  = val
            if key == 'debugToFile':    self.b_debugToFile  = val

        self.dp                 = debug(verbosity   = 0,
                                        level       = -1,
                                        debugFile   = self.str_debugFile,
                                        debugToFile = self.b_debugToFile)

        threading.Thread.__init__(self)


    def run(self):

        timeout = 1
        loop    = 10

        """ Main execution. """

        # Spawn the crunner object container
        self.crunner  = Crunner(cmd         = self.str_cmd,
                                debugToFile = self.b_debugToFile,
                                debugFile   = self.str_debugFile)
        self.crunner.start()

        b_jobsAllDone   = False

        while not b_jobsAllDone:
            try:
                b_jobsAllDone = self.crunner.queueAllDone.get_nowait()
            except queue.Empty:
                # We basically propagate the queue contents "up" the chain.
                self.dp.qprint('Waiting on start job info')
                self.queueStart.put(self.crunner.queueStart.get())

                self.dp.qprint('Waiting on end job info')
                self.queueEnd.put(self.crunner.queueEnd.get())

        self.queueAllDone.put(b_jobsAllDone)
        self.dp.qprint("done with Poller.run")

class Crunner(threading.Thread):
    """
    The wrapper thread about the actual process.
    """

    def __init__(self, **kwargs):
        self.__name             = "Crunner"

        self.queueStart         = queue.Queue()
        self.queueEnd           = queue.Queue()
        self.queueAllDone       = queue.Queue()

        self.str_cmd            = ""

        # Debug parameters
        self.str_debugFile      = '/dev/null'
        self.b_debugToFile      = True

        for k,v in kwargs.items():
            if k == 'cmd':          self.str_cmd        = v
            if k == 'debugFile':    self.str_debugFile  = v
            if k == 'debugToFile':  self.b_debugToFile  = v

        self.shell              = crunner(  verbosity   = 0,
                                            level       = -1,
                                            debugToFile = self.b_debugToFile,
                                            debugFile   = self.str_debugFile)

        self.dp                 = debug(    verbosity   = 0,
                                            level       = -1,
                                            debugFile   = self.str_debugFile,
                                            debugToFile = self.b_debugToFile)
        self.dp.qprint('starting crunner...')

        threading.Thread.__init__(self)

    def jsonJobInfo_queuePut(self, **kwargs):
        """
        Get and return the job dictionary as a json string.
        """

        str_queue   = 'startQueue'
        for k,v in kwargs.items():
            if k == 'queue':    str_queue   = v

        if str_queue == 'startQueue':   queue   = self.queueStart
        if str_queue == 'endQueue':     queue   = self.queueEnd

        # self.dp.qprint(self.shell.d_job)

        queue.put(self.shell.d_job.copy())

    def run(self):

        timeout = 1
        loop    = 10

        """ Main execution. """
        self.dp.qprint("running...")
        self.shell(self.str_cmd)
        # self.shell.jobs_loopctl(    onJobStart  = 'self.jsonJobInfo_queuePut(queue="startQueue")',
        #                             onJobDone   = 'self.jsonJobInfo_queuePut(queue="endQueue")')
        self.shell.jobs_loopctl(    onJobStart  = partial(self.jsonJobInfo_queuePut, queue="startQueue"),
                                    onJobDone   = partial(self.jsonJobInfo_queuePut, queue="endQueue"))
        self.queueAllDone.put(True)
        self.queueStart.put({'allJobsStarted': True})
        self.queueEnd.put({'allJobsDone': True})
        # self.shell.exitOnDone()
        self.dp.qprint('Crunner.run() returning...')
